from abc import ABC, abstractmethod

from simses.commons.state.technology.electrolyzer import ElectrolyzerState


class ElectrolyzerStackModel(ABC):

    def __init__(self):
        super().__init__()

    def update(self, power: float, state: ElectrolyzerState) -> None:
        """
        Updates hydrogen states that are corrolated with the electrolyzer such as current, voltage, hydrogen production,
        water use and temperature

        Parameters
        ----------
        state :

        Returns
        -------

        """
        self.calculate(power, state)
        state.current = self.get_current()
        state.current_density = self.get_current_density()
        state.voltage = self.get_voltage()
        state.hydrogen_production = self.get_hydrogen_production()
        state.oxygen_production = self.get_oxygen_production()
        state.water_use = self.get_water_use()
        state.part_pressure_h2 = self.get_partial_pressure_h2()
        state.part_pressure_o2 = self.get_partial_pressure_o2()
        state.sat_pressure_h2o = self.get_sat_pressure_h2o()

    @abstractmethod
    def calculate(self, power: float, state: ElectrolyzerState) -> None:
        """
        Calculates current, voltage and hydrogen generation based on input power

        Parameters
        ----------
        power : input power in W
        temperature: temperature of electrolyzer in K
        pressure_anode: relative pressure on anode side of electrolyzer in barg (relative to 1 bar)
        pressure_cathode: relative pressure on cathode side of electrolyzer in barg (relative to 1 bar)

        Returns
        -------

        """
        pass

    @abstractmethod
    def get_reference_voltage_eol(self, resistance_increase: float, exchange_currentdensity_decrease: float) -> float:
        """
        return voltage at defined operation point for state of degradation

        :return:
        """
        pass

    @abstractmethod
    def get_current(self) -> float:
        """
        return electrical current of the electrolyzer stack in A

        Returns
        -------

        """
        pass

    @abstractmethod
    def get_current_density(self) -> float:
        """
        return electrical current of the electrolyzer stack in A cm-2

        Returns
        -------

        """
        pass

    @abstractmethod
    def get_voltage(self) -> float:
        """
        Return of electrical voltage of electrolyzer stack in V

        Returns
        -------

        """
        pass

    @abstractmethod
    def get_hydrogen_production(self) -> float:
        """
        Return of total hydrogen generation of the stack in mol/s

        Returns
        -------

        """
        pass

    @abstractmethod
    def get_oxygen_production(self) -> float:
        """
        Return of total oxygen generation of the stack in mol/s

        Returns
        -------

        """
        pass

    @abstractmethod
    def get_water_use(self) -> float:
        """
        Return of water use of electrolyzer stack at anode side

        Returns
        -------

        """
        pass

    @abstractmethod
    def get_number_cells(self) -> float:
        """
        Returns number of serial electrolyseur cells
        -------

        """
        pass

    @abstractmethod
    def get_geom_area_stack(self) -> float:
        """
        Returns geometric area of one cell
        -------

        """
        pass

    @abstractmethod
    def get_nominal_stack_power(self) -> float:
        """
        Returns nominal_stack_power of electrolyzer
        -------

        """
        pass

    @abstractmethod
    def get_heat_capacity_stack(self) -> float:
        """
        Returns nominal_stack_power of electrolyzer
        -------

        """
        pass

    @abstractmethod
    def get_thermal_resistance_stack(self) -> float:
        """
        Returns thermal resistance of electrolyzer
        -------

        """
        pass

    @abstractmethod
    def get_partial_pressure_h2(self) -> float:
        """
        Returns partial pressure of hydrogen at cathode side of electrolyzer
        -------

        """
        pass

    @abstractmethod
    def get_partial_pressure_o2(self) -> float:
        """
        Returns partial pressure of oxigen at anode side of electrolyzer
        -------

        """
        pass

    @abstractmethod
    def get_sat_pressure_h2o(self) -> float:
        """
        Returns staturation pressure of h2o for given temperature
        -------

        """
        pass

    @abstractmethod
    def get_water_in_stack(self) -> float:
        """
        Returns amount of water that is present in the stack
        -------

        """

    @abstractmethod
    def get_nominal_current_density(self) -> float:
        pass

    @abstractmethod
    def close(self) -> None:
        pass



from simses.commons.log import Logger
from simses.system.auxiliary.pump.abstract_pump import Pump


class FixEtaCentrifugalPump(Pump):
    """FixEtaCentrifugalPump is a pump with a constant efficiency."""

    def __init__(self, efficiency: float):
        super().__init__()
        self.__log: Logger = Logger(type(self).__name__)
        self.__eta_pump = efficiency  # pump efficiency p.u.
        if self.__eta_pump <= 0 or self.__eta_pump > 1:
            self.__log.error('Pump efficiency has to be greater than 0 and smaller or equal to 1. It is set to 1.')
            self.__eta_pump = 1
        self.__power = 0.0

    def calculate_pump_power(self, pressure_loss: float) -> None:
        if pressure_loss < 0:
            self.__log.error('Pressure losses are negative. Check sign of pressure drop and flow rate.')
        self.__power = pressure_loss / self.__eta_pump

    def set_eta_pump(self, flow_rate, flow_rate_max, flow_rate_min) -> None:
        pass

    def get_pump_power(self) -> float:
        return self.__power

    def close(self):
        super().close()
        self.__log.close()



import numpy
import scipy.interpolate

from simses.commons.config.data.battery import BatteryDataConfig
from simses.commons.config.simulation.battery import BatteryConfig
from simses.commons.state.technology.lithium_ion import LithiumIonState
from simses.commons.utils.cell_reader import IseaCellReader
from simses.technology.lithium_ion.cell.format.round_18650 import RoundCell18650
from simses.technology.lithium_ion.cell.thermal.default import DefaultThermalCellProperties
from simses.technology.lithium_ion.cell.type import CellType


class IseaCellType(CellType):

    # TODO
    # 1) Make cell type only depend on values from cell reader, no hardcoded values anymore
    # -done- 2) How to handle multiple input files? How to instantiate multiple cell types?
    #   a) One distinct cell type for each file --- drawback: very inconvenient for > 100 cell files
    #   b) Instantiate all cell types in one directory --- how to make this configurable?
    # -> c) file name handed over in constructor of cell type --- batch processing sets files, config could be tricky
    # 3) Documentation is missing
    # 4) Tests are missing

    __EXTENSION: str = '.xml'

    def __init__(self, voltage: float, capacity: float, soh: float, battery_config: BatteryConfig,
                 battery_data_config: BatteryDataConfig, file_name: str):
        self.__name: str = type(self).__name__ + ' with ' + file_name
        cell: IseaCellReader = IseaCellReader(battery_data_config.isea_cell_dir + file_name + self.__EXTENSION)
        isea_cell_properties = cell.get_cell_properties()
        super().__init__(voltage, capacity, soh, isea_cell_properties, DefaultThermalCellProperties(),
                         RoundCell18650(), battery_config)

        ocv_mat: numpy.ndarray = cell.get_open_cicuit_voltage()
        ocv_soc_arr: numpy.ndarray = cell.get_open_cicuit_voltage_soc()
        ocv_temp_arr: numpy.ndarray = cell.get_open_cicuit_voltage_temperature()

        if len(ocv_temp_arr.T) == 1:
            self.__open_circuit_voltage_interpolation = scipy.interpolate.interp1d(ocv_soc_arr, ocv_mat,kind='linear')
            self.__temperature_dependency = False
        else:
            self.__temperature_dependency = True
            self.__open_circuit_voltage_interpolation = scipy.interpolate.interp2d(ocv_soc_arr, ocv_temp_arr.T, ocv_mat,
                                                                                   kind='linear')
        rint_mat: numpy.ndarray = cell.get_internal_resistance()
        rint_soc_arr: numpy.ndarray = cell.get_internal_resistance_soc()
        rint_temp_arr: numpy.ndarray = cell.get_internal_resistance_temperature()
        if len(ocv_temp_arr.T) == 1:
            self.__internal_resistance_interpolation = scipy.interpolate.interp1d(rint_soc_arr, rint_mat,kind='linear')
        else:
            self.__internal_resistance_interpolation = scipy.interpolate.interp2d(rint_soc_arr, rint_temp_arr.T,
                                                                                  rint_mat,kind='linear')

    def get_open_circuit_voltage(self, battery_state: LithiumIonState) -> float:
        soc: float = battery_state.soc * 100.0
        temp: float = battery_state.temperature - 273.15
        if self.__temperature_dependency:
            res: float = float(self.__open_circuit_voltage_interpolation(soc, temp))
        else:
            res: float = float(self.__open_circuit_voltage_interpolation(soc))

        # print(soc, temp, res)
        return res * self.get_serial_scale()

    def get_internal_resistance(self, battery_state: LithiumIonState) -> float:
        soc: float = battery_state.soc * 100.0
        temp: float = battery_state.temperature - 273.15
        if self.__temperature_dependency:
            res: float = float(self.__internal_resistance_interpolation(soc, temp))
        else:
            res: float = float(self.__internal_resistance_interpolation(soc))
        # print(soc, temp, res)
        return res / self.get_parallel_scale() * self.get_serial_scale()

    def get_name(self) -> str:
        return self.__name

    def close(self):
        pass



import numpy as np

from simses.commons.log import Logger
from simses.commons.utils.utilities import check
from simses.system.power_electronics.acdc_converter.abstract_acdc_converter import AcDcConverter


class BonfiglioliAcDcConverter(AcDcConverter):
    """
    Efficiency fit for converter RPS TL-4Q by Bonfiglioli
        (http://www.docsbonfiglioli.com/pdf_documents/catalogue/VE_CAT_RTL-4Q_STD_ENG-ITA_R00_5_WEB.pdf)
    using field data and Notton-Fit as described in master's thesis by Felix Müller.
    """

    __USE_FIELD_DATA = False
    # Notton-Fit parameter
    if __USE_FIELD_DATA:
        # measured data from battery systems providing fcr
        # Charge
        __P0_to_dc = 0.00195
        __K_to_dc = 0.01349
        __min_eff_cha = 0.3441
        # Discharge
        __P0_to_ac = 0.00292
        __K_to_ac = 0.03609
        __min_eff_dis = 0.2742
    else:
        # measured data under testing condition
        # Charge
        __P0_to_dc = 0.0072
        __K_to_dc = 0.034
        __min_eff_cha = 0.5813
        # Discharge
        __P0_to_ac = 0.0072
        __K_to_ac = 0.034
        __min_eff_dis = 0.5813

    __VOLUMETRIC_POWER_DENSITY = 1.667 * 1e5  # W / m3
    # Calculated Value: Nominal Power / (W * H * D)
    __GRAVIMETRIC_POWER_DENSITY = 342  # W/kg
    # Calculated Value: Nominal Power / Weight
    __SPECIFIC_SURFACE_AREA = 2.311 * 1e-5  # in m2 / W
    # Calculated Value: (W*H + W*D + H*D)*2 / Nominal Power

    def __init__(self, max_power: float):
        super().__init__(max_power)
        self.__log: Logger = Logger(type(self).__name__)

    def to_ac(self, power: float, voltage: float) -> float:
        check(power)
        res: float = 0.0
        if power < 0.0:
            res = power / self.__get_efficiency_to_ac(power)
        return res

    def to_dc(self, power: float, voltage: float) -> float:
        check(power)
        res: float = 0.0
        if power > 0.0:
            res = power * self.__get_efficiency_to_dc(power)
        return res

    def to_dc_reverse(self, dc_power: float, voltage: float) -> float:
        check(dc_power)
        res: float = 0.0
        if dc_power > 0.0:
            p = - dc_power / (1 - dc_power * self.__K_to_dc / self.max_power)
            q = - self.__P0_to_dc * dc_power / (
                        1 / self.max_power - abs(dc_power) * self.__K_to_dc / self.max_power ** 2)
            res = max(0.0, -p / 2 + np.sqrt((p / 2) ** 2 - q))
        return res

    def to_ac_reverse(self, dc_power: float, voltage: float) -> float:
        check(dc_power)
        res: float = 0.0
        if dc_power < 0.0:
            p = self.max_power / self.__K_to_ac
            q = (self.__P0_to_ac * self.max_power ** 2 - abs(dc_power) * self.max_power) / self.__K_to_ac
            ac_power = min(0.0, -(-p / 2 + np.sqrt((p / 2) ** 2 - q)))
            res = min(ac_power, self.__min_eff_dis * dc_power)
        return res

    def __get_efficiency_to_ac(self, power) -> float:
        pf = abs(power) / self.max_power
        if 0.0 > pf > 1.0:
            raise Exception('Power factor is not possible: ' + str(pf))
        return max(self.__min_eff_dis, pf / (pf + self.__P0_to_ac + self.__K_to_ac * pf ** 2))

    def __get_efficiency_to_dc(self, power) -> float:
        pf = abs(power) / self.max_power
        if 0.0 > pf > 1.0:
            raise Exception('Power factor is not possible: ' + str(pf))
        return max(self.__min_eff_cha, pf / (pf + self.__P0_to_dc + self.__K_to_dc * pf ** 2))

    @property
    def volume(self) -> float:
        return self.max_power / self.__VOLUMETRIC_POWER_DENSITY

    @property
    def mass(self) -> float:
        return self.max_power / self.__GRAVIMETRIC_POWER_DENSITY

    @property
    def surface_area(self) -> float:
        return self.max_power * self.__SPECIFIC_SURFACE_AREA

    @classmethod
    def create_instance(cls, max_power: float, power_electronics_config=None):
        return BonfiglioliAcDcConverter(max_power)

    def close(self) -> None:
        self.__log.close()



from configparser import ConfigParser
from datetime import datetime

import pytest
from pytz import timezone

from simses.commons.data.no_data_handler import NoDataHandler
from simses.commons.state.technology.lithium_ion import LithiumIonState
from simses.technology.lithium_ion.battery import LithiumIonBattery
from simses.technology.lithium_ion.cell.generic import GenericCell
from simses.technology.lithium_ion.cell.lfp_sony import SonyLFP

THRESHOLD: float = 1e-1

voltage: float = 600  # V
capacity: float = 60e3  # Wh
soh: float = 1.0  # p.u.

start_soc: float = 0.5  # p.u.
start_temperature: float = 300  # K

max_current: float = 205  # A (for Generic Cell in current setup)

time_step: float = 60  # s

start_date: str = '2014-01-01 00:00:00'
date: datetime = datetime.strptime(start_date, '%Y-%m-%d %H:%M:%S')
start_time: float = timezone('UTC').localize(date).timestamp()

battery_config: ConfigParser = ConfigParser()
battery_config.add_section('GENERAL')
battery_config.set('GENERAL', 'START', start_date)
battery_config.set('GENERAL', 'END', '2014-01-02 00:00:00')
battery_config.set('GENERAL', 'TIME_STEP', str(time_step))
battery_config.add_section('BATTERY')
battery_config.set('BATTERY', 'START_SOC', str(start_soc))
battery_config.set('BATTERY', 'MIN_SOC', str(0.0))
battery_config.set('BATTERY', 'MAX_SOC', str(1.0))
battery_config.set('BATTERY', 'EOL', str(0.6))
battery_config.set('BATTERY', 'START_SOH', str(soh))


@pytest.fixture()
def uut(cell: str) -> LithiumIonBattery:
    return LithiumIonBattery(cell=cell, voltage=voltage, capacity=capacity, soc=start_soc, soh=soh,
                             data_export=NoDataHandler(), temperature=start_temperature, storage_id=1, system_id=1,
                             config=battery_config)


@pytest.mark.parametrize('cell, current, result',
                         [
                            (SonyLFP.__name__, 10, 10),
                            (GenericCell.__name__, 10, 10),
                            (GenericCell.__name__, 10, 10)
                         ]
                         )
def test_current(uut: LithiumIonBattery, current: float, result: float):
    uut.distribute_and_run(start_time + time_step, current, voltage)
    battery_state: LithiumIonState = uut.state
    assert abs(battery_state.current - result) <= THRESHOLD



from configparser import ConfigParser

from simses.commons.config.data.data_config import DataConfig


class AuxiliaryDataConfig(DataConfig):

    def __init__(self, path: str = None, config: ConfigParser = None):
        super().__init__(path, config)
        self.__section: str = 'AUXILIARY_DATA'

    @property
    def auxiliary_pump_data_dir(self) -> str:
        """Returns directory of pump data files"""
        return self.get_data_path(self.get_property(self.__section, 'AUXILIARY_PUMP_DATA_DIR'))

    @property
    def pump_eta_file(self) -> str:
        """Returns filename for efficiency file of chosen pump"""
        return self.auxiliary_pump_data_dir + self.get_property(self.__section, 'PUMP_ETA_FILE')



from simses.commons.config.simulation.battery import BatteryConfig
from simses.commons.log import Logger
from simses.commons.state.technology.lithium_ion import LithiumIonState
from simses.technology.lithium_ion.cell.electric.properties import ElectricalCellProperties
from simses.technology.lithium_ion.cell.format.abstract import CellFormat
from simses.technology.lithium_ion.cell.format.prismatic import PrismaticCell
from simses.technology.lithium_ion.cell.thermal.properties import ThermalCellProperties
from simses.technology.lithium_ion.cell.type import CellType


class Samsung78AhNMC(CellType):
    """Characterisation using field data: Master Thesis by Felix Müller"""

    r""" Source:
        Field data from fcr storage system. Datasheet can be found here:
        ...\simses\simulation\storage_system\technology\lithium_ion\cell_type\data\nmc_samsung78Ah_datasheet.pdf
    """

    __CELL_VOLTAGE = 3.7  # V
    __CELL_CAPACITY = 73.95  # Ah
    __CELL_RESISTANCE = 0.636847061 * 10 ** (-3)  # Ohm
    __MAX_VOLTAGE: float = 4.1  # V
    __MIN_VOLTAGE: float = 2.7  # V
    __MIN_TEMPERATURE: float = 248.15  # K
    __MAX_TEMPERATURE: float = 323.15  # K
    __MAX_CHARGE_RATE: float = 2.0  # 1/h
    __MAX_DISCHARGE_RATE: float = 4.0  # 1/h
    __SELF_DISCHARGE_RATE: float = 0.0  # X.X%-soc per day, e.g., 0.015 for 1.5% SOC loss per day
    __MASS: float = 2.0  # kg per cell
    __SPECIFIC_HEAT: float = 823  # J/kgK
    __CONVECTION_COEFFICIENT: float = 15  # W/m2K

    __COULOMB_EFFICIENCY: float = 0.9843  # p.u.

    # Values from nmc_sanyo_ur18650e (source: https://akkuplus.de/Panasonic-UR18650E-37-Volt-2150mAh-Li-Ion-EOL)

    __HEIGHT: float = 125.7  # mm
    __WIDTH: float = 45.6  # mm
    __LENGTH: float = 173.9  # mm

    __ELECTRICAL_PROPS: ElectricalCellProperties = ElectricalCellProperties(__CELL_VOLTAGE, __CELL_CAPACITY,
                                                                            __MIN_VOLTAGE, __MAX_VOLTAGE,
                                                                            __MAX_CHARGE_RATE, __MAX_DISCHARGE_RATE,
                                                                            __SELF_DISCHARGE_RATE, __COULOMB_EFFICIENCY)
    __THERMAL_PROPS: ThermalCellProperties = ThermalCellProperties(__MIN_TEMPERATURE, __MAX_TEMPERATURE, __MASS,
                                                                   __SPECIFIC_HEAT, __CONVECTION_COEFFICIENT)
    __CELL_FORMAT: CellFormat = PrismaticCell(__HEIGHT, __WIDTH, __LENGTH)

    # Parameters for Curve Fitting: Discharge
    __p1_dch = 20.290042045
    __p2_dch = -60.088705239
    __p3_dch = 65.004615460
    __p4_dch = -30.204504426
    __p5_dch = 5.208737944
    __p6_dch = 0.473682851
    __p7_dch = 3.394776910

    # Parameters for Curve Fitting: Charge
    __p1_ch = 20.717822026
    __p2_ch = -60.174639133
    __p3_ch = 62.332299803
    __p4_ch = -25.900465062
    __p5_ch = 2.767903886
    __p6_ch = 0.931931308
    __p7_ch = 3.408247703

    # Parameters for Curve Fitting: Rest
    __p1_rest = 5.255925680
    __p2_rest = -14.288201873
    __p3_rest = 11.785819139
    __p4_rest = -0.932995895
    __p5_rest = -2.416565204
    __p6_rest = 1.282337714
    __p7_rest = 3.375062956

    def __init__(self, voltage: float, capacity: float, soh: float, battery_config: BatteryConfig):
        super().__init__(voltage, capacity, soh, self.__ELECTRICAL_PROPS, self.__THERMAL_PROPS, self.__CELL_FORMAT,
                         battery_config)
        self.__log: Logger = Logger(type(self).__name__)

    def get_open_circuit_voltage(self, battery_state: LithiumIonState) -> float:
        soc = battery_state.soc
        if battery_state.is_charge:
            open_circuit_voltage_cell = self.__p1_ch * soc ** 6 + self.__p2_ch * soc ** 5 + self.__p3_ch * soc ** 4 + \
                                        self.__p4_ch * soc ** 3 + self.__p5_ch * soc ** 2 + self.__p6_ch * soc + self.__p7_ch
        else:
            if battery_state.current == 0:
                open_circuit_voltage_cell = self.__p1_rest * soc ** 6 + self.__p2_rest * soc ** 5 + self.__p3_rest * soc ** 4 + \
                                            self.__p4_rest * soc ** 3 + self.__p5_rest * soc ** 2 + self.__p6_rest * soc + self.__p7_rest
            else:
                open_circuit_voltage_cell = self.__p1_dch * soc ** 6 + self.__p2_dch * soc ** 5 + self.__p3_dch * soc ** 4 + \
                                            self.__p4_dch * soc ** 3 + self.__p5_dch * soc ** 2 + self.__p6_dch * soc + self.__p7_dch
        return open_circuit_voltage_cell * self.get_serial_scale()

    def get_internal_resistance(self, battery_state: LithiumIonState) -> float:
        return float(self.__CELL_RESISTANCE / self.get_parallel_scale() * self.get_serial_scale())

    def close(self):
        self.__log.close()



from abc import ABC, abstractmethod

from simses.commons.state.technology.redox_flow import RedoxFlowState


class ElectrochemicalModel(ABC):
    """Model that calculates the current and voltage of the redox flow stack module."""

    def __init__(self):
        super().__init__()

    @abstractmethod
    def update(self, time: float, redox_flow_state: RedoxFlowState) -> None:
        """
        Updating power (if changes due to redox flow management system), current, voltage, power loss and soc of
        redox_flow_state. In the update function the control system requests are implemented.

        Parameters
        ----------
        time : float
            Current simulation time in s.
        redox_flow_state : RedoxFlowState
            Current state of redox_flow.

        Returns
        -------
            None
        """
        pass

    @abstractmethod
    def close(self) -> None:
        """Closing all resources in ElectrochemicalModel."""
        pass



from simses.commons.config.data.battery import BatteryDataConfig
from simses.commons.config.simulation.battery import BatteryConfig
from simses.technology.lithium_ion.cell.nca_panasonic_ncr import PanasonicNCA


class PanasonicNCAGeneric(PanasonicNCA):

    def __init__(self, voltage: float, capacity: float, soh: float, battery_config: BatteryConfig,
                 battery_data_config: BatteryDataConfig):
        super(PanasonicNCAGeneric, self).__init__(voltage, capacity, soh, battery_config, battery_data_config)



import pytest

from simses.commons.profile.power.constant import ConstantPowerProfile


@pytest.fixture()
def uut(power, scaling_factor):
    return ConstantPowerProfile(power=power, scaling_factor=scaling_factor)


@pytest.mark.parametrize('power, scaling_factor, result',
                         [
                             (0, 0, 0),
                             (1, 0, 0),
                             (0, 1, 0),
                             (1, 1, 1),
                             (5, 3, 15)
                         ]
                         )
def test_next(result, uut):
    tstmp = 0
    while tstmp < 5:
        assert uut.next(tstmp) == result
        tstmp += 1



from simses.commons.timeseries.interpolation.interpolation import Interpolation
from simses.commons.timeseries.timevalue import TimeValue


class LinearInterpolation(Interpolation):
    """
    Linear interpolation behaviour
    """

    def __init__(self):
        super().__init__()

    def interpolate(selfself, time: float, recent: TimeValue, last: TimeValue) -> float:
        factor = (time - last.time) / (recent.time - last.time)
        return factor * recent.value + (1 - factor) * last.value

    def close(self) -> None:
        pass



from datetime import datetime

from simses.commons.log import Logger
from simses.commons.state.system import SystemState
from simses.commons.utils.utilities import format_float
from simses.logic.power_distribution.power_distributor import PowerDistributor
from simses.logic.power_distribution.state import PowerDistributorState


class TechnologyBasedPowerDistributor(PowerDistributor):

    """
    TechnologyBasedPowerDistributor distributes the power depend on the current state and storage technology used.
    A priority list for the technology is used for the order of execution. In addition, TechnologyBasedPowerDistributor
    tries to rebalance the SOC of the various storage devices if possible.
    """

    __SOC_REGULAR_MAX: float = 0.75
    __SOC_REGULAR_MIN: float = 0.25
    __SOC_MAX: float = 0.99
    __SOC_MIN: float = 0.01
    __SOC_THRESHOLD: float = 0.05
    __SOC_TARGET: float = 0.52

    __MIN_REBALANCE_POWER: float = 11000  # W

    __DO_LOG: bool = False
    __DO_REBALANCING: bool = True
    __BALANCE_LOWER_PRIORITY: bool = True
    __BALANCE_ONLY_ONE_DIRECTION: bool = True

    def __init__(self, priorities: [str], storage_systems: dict):
        super().__init__()
        self.__log: Logger = Logger(type(self).__name__)
        self.__states: [PowerDistributorState] = list()
        for priority in priorities:
            for system_id, technology in storage_systems.items():
                if technology == priority:
                    self.__states.append(PowerDistributorState(sys_id=system_id, storage_technology=technology))

    def set(self, time: float, states: [SystemState], power) -> None:
        for state in states:
            pd_state: PowerDistributorState = self.__get_pd_state_for(state)
            pd_state.max_charge_power = state.max_charge_power
            pd_state.max_discharge_power = state.max_discharge_power
            pd_state.soc = state.soc
            pd_state.capacity = state.capacity
            pd_state.time_delta = time - state.time
            pd_state.rebalance_charge_power = 0.0
            pd_state.rebalance_discharge_power = 0.0

    def __soc_max_discharge_power(self, state: PowerDistributorState) -> float:
        denergy = max(0.0, state.soc - self.__SOC_TARGET) * state.capacity
        power = denergy / state.time_delta * 3600.0
        return power

    def __soc_max_charge_power(self, state: PowerDistributorState) -> float:
        denergy = max(0.0, self.__SOC_TARGET - state.soc) * state.capacity
        power = denergy / state.time_delta * 3600.0
        return power

    def __get_pd_state_for(self, state: SystemState) -> PowerDistributorState:
        id: int = self.__get_id_from(state)
        for pd_state in self.__states:
            if pd_state.sys_id == id:
                return pd_state
        self.__log.error('No state found for dc system id ' + str(id))

    def __get_id_from(self, state: SystemState) -> int:
        return int(state.get(SystemState.SYSTEM_DC_ID))

    def __is_charge(self, power: float) -> bool:
        return power > 0.0

    def __search_rebalance_partner(self, rebalance_charge_power: dict, rebalance_discharge_power: dict) -> None:
        for pd_state_charge in self.__states:
            priority_charge: int = self.__states.index(pd_state_charge)
            for pd_state_discharge in self.__states:
                priority_discharge: int = self.__states.index(pd_state_discharge)
                if pd_state_charge.sys_id == pd_state_discharge.sys_id:
                    continue
                if self.__BALANCE_ONLY_ONE_DIRECTION:
                    if self.__BALANCE_LOWER_PRIORITY:
                        if priority_discharge >= priority_charge:
                            continue
                    else:
                        if priority_discharge <= priority_charge:
                            continue
                if pd_state_charge.soc + self.__SOC_THRESHOLD < self.__SOC_REGULAR_MIN < pd_state_discharge.soc - self.__SOC_THRESHOLD \
                        or pd_state_charge.soc + self.__SOC_THRESHOLD < self.__SOC_REGULAR_MAX < pd_state_discharge.soc - self.__SOC_THRESHOLD:
                    possible_charge_power: float = rebalance_charge_power[pd_state_charge.sys_id]
                    possible_discharge_power: float = rebalance_discharge_power[pd_state_discharge.sys_id]
                    possible_charge_power = max(0.0, min(possible_charge_power, self.__soc_max_charge_power(pd_state_charge)))
                    possible_discharge_power = max(0.0, min(possible_discharge_power, self.__soc_max_discharge_power(pd_state_discharge)))
                    possible_charge_power -= pd_state_charge.rebalance_charge_power
                    possible_discharge_power -= pd_state_discharge.rebalance_discharge_power
                    possible_recharge: float = max(0.0, min(possible_charge_power, possible_discharge_power))
                    if possible_recharge > self.__MIN_REBALANCE_POWER:
                        pd_state_charge.rebalance_charge_power += possible_recharge
                        pd_state_discharge.rebalance_discharge_power += possible_recharge

    def __get_regular_charge_power(self, state: PowerDistributorState) -> float:
        return 0.0 if state.soc > self.__SOC_REGULAR_MAX else state.max_charge_power

    def __get_max_charge_power(self, state: PowerDistributorState) -> float:
        return 0.0 if state.soc > self.__SOC_MAX else state.max_charge_power

    def __get_regular_discharge_power(self, state: PowerDistributorState) -> float:
        return 0.0 if state.soc < self.__SOC_REGULAR_MIN else state.max_discharge_power

    def __get_max_discharge_power(self, state: PowerDistributorState) -> float:
        return 0.0 if state.soc < self.__SOC_MIN else state.max_discharge_power

    def __get_regular_power_from(self, is_charge: bool, state: PowerDistributorState) -> float:
        if is_charge:
            return self.__get_regular_charge_power(state)
        else:
            return self.__get_regular_discharge_power(state)

    def __get_max_power_from(self, is_charge: bool, state: PowerDistributorState) -> float:
        if is_charge:
            return self.__get_max_charge_power(state)
        else:
            return self.__get_max_discharge_power(state)

    def get_power_for(self, power_target: float, state: SystemState) -> float:
        is_charge: bool = self.__is_charge(power_target)
        remaining_power: float = abs(power_target)
        local_power: float = 0.0
        emergency_power: float = 0.0
        balancing_power: float = 0.0
        id: int = self.__get_id_from(state)
        rebalance_charge_power: dict = dict()
        rebalance_discharge_power: dict = dict()
        for pd_state in self.__states:
            regular_available_power: float = self.__get_regular_power_from(is_charge, pd_state)
            allocation_power: float = max(0.0, min(remaining_power, regular_available_power))
            if pd_state.sys_id == id:
                local_power = allocation_power
                remaining_power -= abs(local_power)
            else:
                remaining_power -= allocation_power
            if is_charge:
                rebalance_charge_power[pd_state.sys_id] = regular_available_power - allocation_power
                rebalance_discharge_power[pd_state.sys_id] = self.__get_regular_discharge_power(pd_state)
            else:
                rebalance_charge_power[pd_state.sys_id] = self.__get_regular_charge_power(pd_state)
                rebalance_discharge_power[pd_state.sys_id] = regular_available_power - allocation_power
        if remaining_power > 0.0:
            for pd_state in self.__states:
                max_available_power: float = self.__get_max_power_from(is_charge, pd_state)
                if pd_state.sys_id == id:
                    emergency_power = max(0.0, min(remaining_power, max_available_power - local_power))
                    remaining_power -= abs(emergency_power)
                else:
                    regular_available_power: float = self.__get_regular_power_from(is_charge, pd_state)
                    remaining_power -= min(remaining_power, max_available_power - regular_available_power)
                rebalance_charge_power[pd_state.sys_id] = 0.0
                rebalance_discharge_power[pd_state.sys_id] = 0.0
        if self.__DO_REBALANCING:
            self.__search_rebalance_partner(rebalance_charge_power, rebalance_discharge_power)
            pd_state: PowerDistributorState = self.__get_pd_state_for(state)
            balancing_power = pd_state.rebalance_charge_power - pd_state.rebalance_discharge_power
            if abs(balancing_power) > 0.0 and self.__DO_LOG:
                self.__log.error('Rebalancing ' + str(id) + ' with ' + str(int(balancing_power/1000.0)) + ' kW, '
                                    'SOC:' + format_float(state.soc) + ' @ ' + str(datetime.fromtimestamp(state.time)))
        local_power += emergency_power
        if is_charge:
            return local_power + balancing_power
        else:
            return -local_power + balancing_power

    def close(self):
        self.__log.close()



import math

import pandas as pd
import scipy.interpolate

from simses.commons.config.data.battery import BatteryDataConfig
from simses.commons.config.simulation.battery import BatteryConfig
from simses.commons.log import Logger
from simses.commons.state.technology.lithium_ion import LithiumIonState
from simses.technology.lithium_ion.cell.electric.properties import ElectricalCellProperties
from simses.technology.lithium_ion.cell.format.abstract import CellFormat
from simses.technology.lithium_ion.cell.format.prismatic import PrismaticCell
from simses.technology.lithium_ion.cell.thermal.properties import ThermalCellProperties
from simses.technology.lithium_ion.cell.type import CellType


class LTOLMO(CellType):
    """An LTO_LMO is a special cell type and inherited by CellType"""

    # Source:
    # Thomas Nemeth, Philipp Schröer, Matthias Kuipers, Dirk Uwe Sauer:
    # Lithium titanate oxide battery cells for high-power automotive applications -
    # Electro-thermal  properties, aging behavior and cost considerations,
    # Journal of Energy Storage 31 (2020) 101656, https://doi.org/10.1016/j.est.2020.101656

    __SOC_HEADER: str = 'SOC'
    __SOC_IDX: int = 0
    __TEMP_IDX: int = 1
    __C_RATE_IDX: int = 0
    __ETA_IDX: int = 1

    __CELL_VOLTAGE: float = 2.4  # V
    __CELL_CAPACITY: float = 2.9  # Ah
    __MAX_VOLTAGE: float = 2.8  # V
    __MIN_VOLTAGE: float = 1.8  # V
    __CELL_RESISTANCE: float = 0.81 * 10 ** (-3)  # Ohm # from lmo_daimler
    __MIN_TEMPERATURE: float = 263.15  # K
    __MAX_TEMPERATURE: float = 318.15  # K
    __MAX_CHARGE_RATE: float = 10.0  # 1/h
    __MAX_DISCHARGE_RATE: float = 10.0  # 1/h
    __SELF_DISCHARGE_RATE: float = 0.0  # X.X%-soc per month in second step, e.g., 0.015 / (30.5 * 24 * 3600) for 1.5%
    __MASS: float = 0.151  # kg per cell
    # Source of specific heat:
    # Hamidreza Behi, Danial Karimi, Mohammadreza Behi, Joris Jaguemont, Morteza Ghanbarpour, Masud Behnia, Maitane Berecibar, Joeri Van Mierlo:
    # Thermal management analysis using heat pipe in the high current discharging of lithium-ion battery in electric vehicles,
    # Journal of Energy Storage 32 (2020) 101893, https://doi.org/10.1016/j.est.2020.101893
    __SPECIFIC_HEAT: float = 1150  # J/kgK
    __CONVECTION_COEFFICIENT: float = 15  # W/m2K convection coefficient of LFP Sony
    __COULOMB_EFFICIENCY: float = 1.0  # p.u.

    __LENGTH: float = 69  # mm
    __WIDE: float = 62  # mm
    __HEIGHT: float = 13  # mm

    __ELECTRICAL_PROPS: ElectricalCellProperties = ElectricalCellProperties(__CELL_VOLTAGE, __CELL_CAPACITY,
                                                                            __MIN_VOLTAGE, __MAX_VOLTAGE,
                                                                            __MAX_CHARGE_RATE, __MAX_DISCHARGE_RATE,
                                                                            __SELF_DISCHARGE_RATE,
                                                                            __COULOMB_EFFICIENCY)
    __THERMAL_PROPS: ThermalCellProperties = ThermalCellProperties(__MIN_TEMPERATURE, __MAX_TEMPERATURE, __MASS,
                                                                   __SPECIFIC_HEAT, __CONVECTION_COEFFICIENT)
    __CELL_FORMAT: CellFormat = PrismaticCell(__HEIGHT, __WIDE, __LENGTH)

    def __init__(self, voltage: float, capacity: float, soh: float, battery_config: BatteryConfig,
                 battery_data_config: BatteryDataConfig):
        super().__init__(voltage, capacity, soh, self.__ELECTRICAL_PROPS, self.__THERMAL_PROPS, self.__CELL_FORMAT,
                         battery_config)
        self.__log: Logger = Logger(type(self).__name__)

        # Physical parameters for lithium_ion thermal model
        # See SimSES Paper for Sources
        self.__volume = self.__LENGTH * self.__WIDE * self.__HEIGHT * 10 ** (-9) \
                        * self.get_serial_scale() * self.get_parallel_scale()  # m3
        self.__surface_area = 2 * (self.__LENGTH * self.__WIDE + self.__LENGTH * self.__HEIGHT + self.__HEIGHT * self.__WIDE) \
                              * 10 ** (-6) * self.get_serial_scale() * self.get_parallel_scale()  # m2



    def get_open_circuit_voltage(self, battery_state: LithiumIonState) -> float:
        '''Parameters build with ocv fitting'''
        a1 = -98.0995
        a2 = 0.6497
        a3 = -17.7261
        a4 = -3.1726
        b1 = 1.0308
        b2 = -3.3736
        k0 = 2.6329
        k1 = -1.4900
        k2 = -4.8486
        k3 = 1.5915
        k4 = -0.4285
        k5 = 0.0033

        soc = battery_state.soc

        ocv = k0 + \
              k1 / (1 + math.exp(a1 * (soc - b1))) + \
              k2 / (1 + math.exp(a2 * (soc - b2))) + \
              k3 / (1 + math.exp(a3 * (soc - 1))) +\
              k4 / (1 + math.exp(a4 * soc)) +\
              k5 * soc

        return ocv * self.get_serial_scale()

    def get_internal_resistance(self, battery_state: LithiumIonState) -> float:
        return float(self.__CELL_RESISTANCE / self.get_parallel_scale() * self.get_serial_scale())

    def close(self):
        self.__log.close()


from abc import ABC, abstractmethod

from simses.commons.state.technology.redox_flow import RedoxFlowState
from simses.system.auxiliary.pump.abstract_pump import Pump


class PumpAlgorithm(ABC):

    def __init__(self, pump: Pump):
        self.__pump: Pump = pump
        self.__soc_start_time_step = 0.5

    def update(self, redox_flow_state: RedoxFlowState, soc_start_time_step: float) -> None:
        """
        Updates flow rate and pressure drop of the current redox_flow_state and starts the calculation of
        the pump_power.

        Parameters
        ----------
        redox_flow_state : RedoxFlowState
            Current state of redox_flow.
        soc_start_time_step: float
            SOC at the begin of the time step in p.u.

        Returns
        -------

        """
        flow_rate_min = self.get_flow_rate_min()
        flow_rate_max = self.get_flow_rate_max()
        self.__soc_start_time_step = soc_start_time_step
        redox_flow_state.pressure_drop_catholyte = self.get_pressure_drop_catholyte(redox_flow_state)
        redox_flow_state.pressure_drop_anolyte = self.get_pressure_drop_anolyte(redox_flow_state)
        redox_flow_state.flow_rate_catholyte = self.get_flow_rate_catholyte(redox_flow_state)
        redox_flow_state.flow_rate_anolyte = self.get_flow_rate_anolyte(redox_flow_state)
        redox_flow_state.pressure_loss_catholyte = (redox_flow_state.flow_rate_catholyte *
                                                    redox_flow_state.pressure_drop_catholyte)
        redox_flow_state.pressure_loss_anolyte = (redox_flow_state.flow_rate_anolyte *
                                                  redox_flow_state.pressure_drop_anolyte)

        self.__pump.set_eta_pump(redox_flow_state.flow_rate_catholyte, flow_rate_max, flow_rate_min)
        self.__pump.calculate_pump_power(redox_flow_state.pressure_loss_catholyte)
        pump_power_catholyte = self.__pump.get_pump_power()

        self.__pump.set_eta_pump(redox_flow_state.flow_rate_anolyte, flow_rate_max, flow_rate_min)
        self.__pump.calculate_pump_power(redox_flow_state.pressure_loss_anolyte)
        pump_power_anolyte = self.__pump.get_pump_power()

        redox_flow_state.pump_power = pump_power_catholyte + pump_power_anolyte

    @abstractmethod
    def get_pressure_drop_anolyte(self, redox_flow_state: RedoxFlowState) -> float:
        """
        Determines the pressure drop over the stack module and pipe system for the anolyte side in Pa.

        Parameters
        ----------
        redox_flow_state : RedoxFlowState
            Current state of redox_flow.

        Returns
        -------
        float :
            Pressure drop of the anolyte side in Pa.
        """
        pass

    @abstractmethod
    def get_pressure_drop_catholyte(self, redox_flow_state: RedoxFlowState) -> float:
        """
        Determines the pressure drop over the stack module and pipe system for the catholyte side in Pa.

        Parameters
        ----------
        redox_flow_state : RedoxFlowState
            Current state of redox_flow.

        Returns
        -------
        float :
            Pressure drop of the catholyte side in Pa.
        """
        pass

    @abstractmethod
    def get_flow_rate_anolyte(self, redox_flow_state: RedoxFlowState) -> float:
        """
        Determines the the flow rate of the anolyte side in m^3/s.

        Parameters
        ----------
        redox_flow_state : RedoxFlowState
            Current state of redox_flow.

        Returns
        -------
        float :
            Flow rate of the anolyte side in m^3/s.
        """
        pass

    @abstractmethod
    def get_flow_rate_catholyte(self, redox_flow_state: RedoxFlowState) -> float:
        """
        Determines the the flow rate of the catholyte side in m^3/s.

        Parameters
        ----------
        redox_flow_state : RedoxFlowState
            Current state of redox_flow.

        Returns
        -------
        float :
            Flow rate of the catholyte side in m^3/s.
        """
        pass

    @abstractmethod
    def get_flow_rate_max(self) -> float:
        """
        Maximal needed flow rate in the system in m^3/s.

        Returns
        -------
        float:
            Maximal flow rate in m^3/s.
        """
        pass

    @abstractmethod
    def get_flow_rate_min(self) -> float:
        """
        Minimal needed flow rate in the system in m^3/s.

        Returns
        -------
        float:
            Minimal flow rate in m^3/s.
        """
        pass

    def get_soc_begin(self):
        return self.__soc_start_time_step

    @abstractmethod
    def close(self):
        """Closing all resources in PumpAlgorithm."""
        self.__pump.close()



import csv
import datetime
import inspect
import math
import os
import shutil
import sys
import urllib.request
from os.path import basename
from urllib.error import URLError

from simses.commons.error import InfiniteNumberError, NanNumberError, ComplexNumberError


def remove_file(file: str) -> None:
    """
    Removes file

    Parameters
    ----------
    file :
        path to file

    Returns
    -------

    """
    if os.path.isfile(file):
        os.remove(file)


def remove_all_files_from(directory: str) -> None:
    """
    Function to remove all files from a directory

    Parameters
    ----------
    directory : folder path

    Returns
    -------

    """
    try:
        for item in os.listdir(directory):
            file = os.path.join(directory, item)
            remove_file(file)
    except FileNotFoundError:
        # print(directory + ' does not exist')
        pass


def copy_all_files(source: str, target: str):
    """
    Function to copy all files in a new folder

    Parameters
    ----------
    source : path (string) to the source folder
    target : path (string) to the target folder

    Returns
    -------

    """

    for item in os.listdir(source):
        path = os.path.join(source, item)
        if os.path.isfile(path):
            shutil.copy(path, target)


def download_file(url_path: str, file_name: str) -> None:
    """
    Downloading a file from the given url and stores it to file_name

    Raises FileNotFoundError if file at URL is not available

    Parameters
    ----------
    url_path :
        URL to file to download
    file_name :
        path where file should be stored

    Returns
    -------

    """
    try:
        with urllib.request.urlopen(url_path) as response:
            print('Downloading ' + basename(file_name) + ' from ' + url_path + ' ...')
            with open(file_name, 'wb') as out_file:
                shutil.copyfileobj(response, out_file)
    except URLError as err:
        raise FileNotFoundError(err, file_name + ' could not be found on server.')


def create_directory_for(path: str, warn: bool = False) -> None:
    """
    Function to create a folder at a specific path

    Parameters
    ----------
    path : str

    Returns
    -------

    """
    if not os.path.exists(path):
        os.makedirs(path)
    elif warn:
        sys.stderr.write('WARNING: ' + path + ' exists already')
        sys.stderr.flush()


def format_float(value: float, decimals: int = 2) -> str:
    """
    Formatting value to string with given decimals

    Parameters
    ----------
    value :
        given value
    decimals :
        round to number of decimals

    Returns
    -------
    str:
        value as string
    """
    return f"{value:.{decimals}f}"


def write_to_file(filename: str, _list: [str], append: bool = False) -> None:
    """
    Writes given list of strings to file

    Parameters
    ----------
    filename :
        path to file
    _list :
        list with values as strings
    append :
        flag to append to file or (over)write file

    Returns
    -------
    None:
        None
    """
    with open(filename, 'a' if append else 'w', newline='') as writeFile:
        writer = csv.writer(writeFile)
        writer.writerow(_list)
    writeFile.close()


def all_non_abstract_subclasses_of(cls, exclude: list=list()) -> list:
    """
    Generates a list of non-abstract subclass from given class

    Parameters
    ----------
    exclude :
        list of classes which should be excluded from result list
    cls :
        given class

    Returns
    -------
    list:
        list with non-abstract subclasses
    """
    res = list()
    subs = set(cls.__subclasses__()).union([s for c in cls.__subclasses__() for s in all_non_abstract_subclasses_of(c)])
    for sub in subs:
        if not inspect.isabstract(sub) and sub not in exclude:
            res.append(sub)
    return res


def search_path_for_file_in(directory: str, filename: str) -> str:
    """
    Searching path where file with filename is located within all subdirectories of directory

    Parameters
    ----------
    directory :
        base directory for search
    filename :
        searched filename

    Returns
    -------
    str:
        path where filename is located, if filename was found - empty string otherwise

    """
    path: str = ''
    for item in os.listdir(directory):
        file = os.path.join(directory, item)
        if os.path.isfile(file) and file.endswith(filename):
            if not path:
                path = directory
                break
        if os.path.isdir(file):
            local_dir = search_path_for_file_in(file, filename)
            if not path and local_dir:
                path = local_dir
                break
    return path


def get_path_for(filename: str, max_depth: int = 4) -> str:
    """
    Searching path for filename starting with current working directory and all of its subdirectories. If file was not
    found, it goes up the parent directories of the CWD with a given maxium depth

    Parameters
    ----------
    filename :
        search filename
    max_depth :
        maximum depth of parent directories in which should be searched, default: 3

    Returns
    -------
    str:
        path where filename is located or a FileNotFoundError

    """
    res: str = ''
    depth: int = 0
    path: str = os.path.dirname(os.path.abspath(__file__))
    while not res and depth < max_depth:
        res = search_path_for_file_in(path, filename)
        path = os.path.dirname(path)
        depth += 1
    if not res:
        raise FileNotFoundError(filename + ' could not be found')
    return res + '/'


def convert_path_codec(path: str, encode: str = 'latin-1', decode: str = 'utf-8') -> str:
    """
    Enconding and decoding path with given codecs, e.g. in order to convert German Umlaut

    Parameters
    ----------
    path :
        string which should be converted
    encode :
        enconding codec, defaults: latin-1
    decode :
        decoding codec, defaults: utf-8

    Returns
    -------
    str:
        decoded path

    """
    try:
        if is_codec(path, decode):
            return path
        return path.encode(encode).decode(decode)
    except (UnicodeEncodeError, UnicodeDecodeError) as err:
        sys.stderr.write(err)
        sys.stderr.flush()
        return path


def is_codec(string: str, codec: str = 'utf-8') -> bool:
    """
    Checks if given string is encoded as given codec

    Parameters
    ----------
    string :
        string to check codec
    codec :
        codec which should be checked, default: utf-8

    Returns
    -------
    bool:
        True if string is encoded with codec, False otherwise
    """
    try:
        string.encode(codec).decode(codec)
        return True
    except (UnicodeEncodeError, UnicodeDecodeError):
        return False


def add_month_to(date: datetime.datetime) -> datetime.datetime:
    """
    Generates a new datetime object from given datetime object with a month added.
    If month is last month of year, it returns first month of next year.

    Parameters
    ----------
    date :
        given datetime object

    Returns
    -------
    datetime:
        new datetime object
    """
    if date.month == 12:
        date = date.replace(year=date.year + 1, month=1)
    else:
        date = date.replace(month=date.month + 1)
    return date


def add_year_to(date: datetime.datetime) -> datetime.datetime:
    """
    Generates a new datetime object from given datetime object with a year added.

    Parameters
    ----------
    date :
        given datetime object

    Returns
    -------
    datetime:
        new datetime object
    """
    return date.replace(year=date.year + 1)


def get_year_from(tstmp: float) -> float:
    """
    Calculates year from given timestamp

    Parameters
    ----------
    tstmp :
        given timestamp in epoch time

    Returns
    -------
    float :
        year of given epoch timestamp
    """
    return datetime.datetime.fromtimestamp(tstmp).year


def get_maximum_from(values: list) -> float:
    """
    Gets maximum value from given values

    Parameters
    ----------
    values :
        list of values

    Returns
    -------
    float:
        maximum value from given list
    """
    if values:
        max_value: float = max(values)
    else:
        max_value: float = 0.0
    return max_value


def get_average_from(values: list) -> float:
    """
    Calculates average from given value list

    Parameters
    ----------
    values :
        list of values

    Returns
    -------
    float:
        average from values
    """
    if values:
        average_value: float = sum(values) / len(values)
    else:
        average_value: float = 0.0
    return average_value


def check(value: float) -> None:
    if math.isinf(value):
        raise InfiniteNumberError()
    if math.isnan(value):
        raise NanNumberError()
    if isinstance(value, complex):
        raise ComplexNumberError()



import pandas as pd

from simses.commons.config.data.fuel_cell import FuelCellDataConfig
from simses.commons.constants import Hydrogen
from simses.commons.log import Logger
from simses.technology.hydrogen.fuel_cell.stack.stack_model import \
    FuelCellStackModel


class JupiterFuelCell(FuelCellStackModel):

    """
    A Jupiter-Fuelcell is a special typ of a PEM-Fuelcell with an open cathode from SFC Energy:
    https://www.efoy-pro.com/efoy-pro/efoy-jupiter-2-5/
    """

    __POWER_HEADER = 'cellpower at 1 bar'
    __VOLTAGE_IDX = 1
    __POWER_IDX = 1
    __CURRENT_IDX = 0
    __MAX_POWER_Cell = 48.9494
    __NUMBER_CELLS = 46  # möglich sind [10,12,13,18,20,28,36,39,46,56,64,66,70,73,80]

    def __init__(self, fuel_cell_maximal_power: float, fuel_cell_data_config: FuelCellDataConfig):
        super(JupiterFuelCell, self).__init__()

        self.__log: Logger = Logger(type(self).__name__)
        self.__SPEC_MAX_POWER = 0.49  # W/cm2
        self.__NOM_STACK_POWER = fuel_cell_maximal_power
        self.__GEOM_AREA = 250  # cm2
        self.__PC_FILE = fuel_cell_data_config.jupiter_fuel_cell_pc_file
        self.__POWER_FILE = fuel_cell_data_config.jupiter_fuel_cell_power_file
        self.__mass_flow_h2 = 0  # mol/s
        self.__MAX_CURRENT_DENSITY = 5  # A/cm2
        self.__MIN_CURRENT_DENSITY = 0  # mA
        self.__HEAT_CAPACITY = Hydrogen.EPS
        self.__polarization_curve = pd.read_csv(self.__PC_FILE)  # V/(mA/cm2) per cell
        self.__power_curve = pd.read_csv(self.__POWER_FILE)  # W/cm2 per cell
        self.__current_arr = self.__polarization_curve.iloc[:, self.__CURRENT_IDX] * self.__GEOM_AREA  # mA
        self.__polarization_arr = self.__polarization_curve.iloc[:, self.__VOLTAGE_IDX] * self.__NUMBER_CELLS  # V
        self.__power_arr = self.__power_curve.iloc[:, self.__POWER_IDX] * self.__GEOM_AREA * self.__NUMBER_CELLS  # W


    def calculate(self, power) -> None:
        power_abs = abs(power)
        power_idx = abs(self.__power_curve[self.__POWER_HEADER]
                        * self.__GEOM_AREA * self.__NUMBER_CELLS - power_abs).idxmin()  # hier wird eine positive Leistung erwartet!!!
        self.__voltage = self.__polarization_curve.iloc[power_idx, self.__VOLTAGE_IDX] * self.__NUMBER_CELLS  # V
        self.__current = - self.__power_curve.iloc[power_idx, self.__CURRENT_IDX] * self.__GEOM_AREA  # mA
        self.__hydrogen_consumption = self.__current / (2 * Hydrogen.FARADAY_CONST) * self.__NUMBER_CELLS  # mol/s
        self.__power = - self.__power_curve.iloc[power_idx, self.__POWER_IDX] * self.__GEOM_AREA * self.__NUMBER_CELLS  # W
        a = 1

    def get_current(self):
        # print('the current is: ' + str(self.__current) + ' mA')
        return self.__current

    def get_voltage(self):
        # print('the voltage is: ' + str(self.__voltage) + ' V')
        return self.__voltage

    def get_hydrogen_consumption(self):
        # print('the massflow of hydrogen ist: ' + str(self.__mass_flow_h2) + ' mol/s')
        return - self.__hydrogen_consumption  # positive value

    def get_nominal_stack_power(self):
        return self.__NOM_STACK_POWER

    # def get_power(self):
    #     return self.__power

    def get_number_cells(self):
        return self.__NUMBER_CELLS

    # def get_max_spec_power(self):
    #     return self.get_max_spec_power()

    def get_geom_area_stack(self):
        return self.__GEOM_AREA

    def get_heat_capactiy_stack(self):
        return self.__HEAT_CAPACITY

    def get_current_density(self) -> float:
        return self.__current / self.__GEOM_AREA

    def close(self):
        self.__log.close()




import pandas as pd
import scipy.interpolate
from simses.commons.config.data.battery import BatteryDataConfig
from simses.commons.state.technology.lithium_ion import LithiumIonState
from simses.technology.lithium_ion.cell.type import CellType
from simses.technology.lithium_ion.degradation.calendar.calendar_degradation import CalendarDegradationModel


class MolicelNMCCalendarDegradationModel(CalendarDegradationModel):

    # Values based on MA Ni Chuanqin (EES, TUM) and adapted from Yulong Zhao in order to have the same structure
    # as in the aging models of Maik Neumann (SonyLFP)

    __SOC_IDX = 0
    __TEMP_IDX = 1
    __LENGTH_TEMP_ARRAY = 40

    def __init__(self, cell_type: CellType, battery_data_config: BatteryDataConfig):
        super().__init__(cell_type)
        self.__rinc_cal = 0
        self.__capacity_loss = 0
        self.__resistance_increase = 0
        self.__initial_capacity = cell_type.get_nominal_capacity()
        self.__capacity_loss_calendar = cell_type.get_calendar_capacity_loss_start()

        capacity_cal_file: str = battery_data_config.nmc_molicel_capacity_cal_file
        capacity_cal = pd.read_csv(capacity_cal_file)  # -
        capacity_cal_mat = capacity_cal.iloc[:(self.__LENGTH_TEMP_ARRAY + 1), 2:]
        soc_arr = capacity_cal.iloc[:, self.__SOC_IDX]
        temp_arr = capacity_cal.iloc[:(self.__LENGTH_TEMP_ARRAY + 1), self.__TEMP_IDX]
        self.__capacity_cal_interp1d = scipy.interpolate.interp2d(soc_arr, temp_arr.T, capacity_cal_mat, kind='linear')

        ri_cal_file: str = battery_data_config.nmc_molicel_ri_cal_file
        ri_cal = pd.read_csv(ri_cal_file)  # -
        ri_cal_mat = ri_cal.iloc[:(self.__LENGTH_TEMP_ARRAY + 1), 2:]
        soc_arr = ri_cal.iloc[:, self.__SOC_IDX]
        temp_arr = ri_cal.iloc[:(self.__LENGTH_TEMP_ARRAY + 1), self.__TEMP_IDX]
        self.__ri_cal_interp1d = scipy.interpolate.interp2d(soc_arr, temp_arr.T, ri_cal_mat, kind='linear')

    def calculate_degradation(self, time: float, battery_state: LithiumIonState) -> None:
        time_passed = time - battery_state.time
        qloss: float = self.__capacity_loss_calendar # only calendar losses

        k_capacity_cal = self.get_stressfkt_ca_cal(battery_state)
        virtual_time = (qloss/k_capacity_cal)**(4/3) # virtual aging time in weeks

        capacity_loss = k_capacity_cal*(virtual_time + time_passed/(86400*7))**0.75 - qloss # pu

        self.__capacity_loss_calendar += capacity_loss  # pu
        self.__capacity_loss = capacity_loss * self.__initial_capacity # Ah

    def calculate_resistance_increase(self, time: float, battery_state: LithiumIonState) -> None:
        time_passed = time - battery_state.time
        rinc_cal = self.__rinc_cal
        k_ri_cal = self.get_stressfkt_ri_cal(battery_state)

        virtual_time = (rinc_cal/k_ri_cal)**2

        resistance_increase = k_ri_cal*(virtual_time + time_passed/(86400*7))**0.5 - rinc_cal
        self.__resistance_increase = resistance_increase  # pu
        self.__rinc_cal += resistance_increase

    def get_degradation(self) -> float:
        return self.__capacity_loss

    def get_resistance_increase(self) -> float:
        return self.__resistance_increase

    def get_stressfkt_ca_cal(self, battery_state: LithiumIonState) -> float:
        """
        get the stress factor for calendar aging capacity loss

        Parameters
        ----------
        battery_state : state including soc and temperature

        Returns
        -------
        float : stress parameters of calendar aging (capacity loss)

        """
        return float(self.__capacity_cal_interp1d(battery_state.soc, battery_state.temperature))

    def get_stressfkt_ri_cal(self, battery_state: LithiumIonState) -> float:
        """
        get the stress factor for calendar aging resistance increase

        Parameters
        ----------
        battery_state : state including soc and temperature

        Returns
        -------
        float : stress parameters of calendar aging (resistance increase)

        """
        return float(self.__ri_cal_interp1d(battery_state.soc, battery_state.temperature))

    def reset(self, battery_state: LithiumIonState) -> None:
        self.__rinc_cal = 0
        self.__capacity_loss = 0
        self.__resistance_increase = 0

    def close(self) -> None:
        pass



from datetime import datetime
from simses.commons.config.generation.generator import ConfigGenerator
from simses.commons.config.simulation.battery import BatteryConfig
from simses.commons.config.simulation.energy_management import EnergyManagementConfig
from simses.commons.config.simulation.general import GeneralSimulationConfig
from simses.commons.config.simulation.profile import ProfileConfig
from simses.commons.config.simulation.simulation_config import SimulationConfig
from simses.commons.config.simulation.system import StorageSystemConfig


class SimulationConfigGenerator(ConfigGenerator):

    """
    The SimulationConfigGenerator is a convenience class for generating a config for a SimSES simulation. Prior knowledge
    of the options and structure of SimSES is recommended. Before using SimSES within another application it is very
    helpful to get to know the concepts of SimSES by using it as a standalone tool.

    This config generator allows the user to focus only on the options to generate as well as the systems to instantiate
    without needing to worry about the config structure and naming. However, names of possible classes to instantiate
    are necessary. Basic implementations like "No"-implementations are provided as convenience methods, other types
    need to be named directly.

    First, you generate options for different kind of components like housing, hvac, acdc / dcdc converter, etc.. For
    all of these methods you get a return key for this kind of component. This key is needed for the instantiation methods.
    Second, you define the AC and DC systems with the keys defined and some other values like max. AC power, capacity and
    intermediate circuit voltage.

    You are able to load configs (defaults, local, or your own config file). Please consider that maybe AC and DC systems
    are already defined which will be instantiated. You can clear these options with the provided clear functions.
    """

    __AC_SYSTEM: str = 'ac_system'
    __HVAC: str = 'hvac'
    __HOUSING: str = 'housing'
    __DCDC_CONVERTER: str = 'dcdc'
    __ACDC_CONVERTER: str = 'acdc'
    __TECHNOLOGY: str = 'technology'

    def __init__(self):
        super(SimulationConfigGenerator, self).__init__()

    def load_default_config(self) -> None:
        """
        Loads defaults config

        Returns
        -------

        """
        path: str = SimulationConfig.CONFIG_PATH
        path += SimulationConfig.CONFIG_NAME
        path += SimulationConfig.DEFAULTS
        self.load_config_from(path)

    def load_local_config(self) -> None:
        """
        Loads local config

        Returns
        -------

        """
        path: str = SimulationConfig.CONFIG_PATH
        path += SimulationConfig.CONFIG_NAME
        path += SimulationConfig.LOCAL
        self.load_config_from(path)

    def load_specific_config(self, config_name: str) -> None:
        """
        Loads specific config given by config-name without ".local.ini" at the end

        Returns
        -------

        """
        path: str = SimulationConfig.CONFIG_PATH
        path += config_name
        path += SimulationConfig.LOCAL
        self.load_config_from(path)


    def __check_time_format(self, time: str) -> None:
        if time is None:
            return
        try:
            datetime.strptime(time, GeneralSimulationConfig.TIME_FORMAT)
        except ValueError:
            raise ValueError('Incorrect time format. Expected time format: ' + GeneralSimulationConfig.TIME_FORMAT)

    def get_time_format(self) -> str:
        """

        Returns
        -------
        str:
            expected time format for simulation start and end
        """
        return GeneralSimulationConfig.TIME_FORMAT

    def set_simulation_time(self, start: str = None, end: str = None, time_step: float = 60.0, loop: int = 1) -> None:
        """
        Setting simulation time parameters

        Parameters
        ----------
        start :
            simulation start in expected format
        end :
            simulation start in expected format
        time_step :
            simulation time step in seconds
        loop :
            looping the given simulation time period

        Returns
        -------

        """
        self.__check_time_format(start)
        self.__check_time_format(end)
        self._set(GeneralSimulationConfig.SECTION, GeneralSimulationConfig.START, start)
        self._set(GeneralSimulationConfig.SECTION, GeneralSimulationConfig.END, end)
        self._set(GeneralSimulationConfig.SECTION, GeneralSimulationConfig.TIME_STEP, str(time_step))
        self._set(GeneralSimulationConfig.SECTION, GeneralSimulationConfig.LOOP, str(loop))

    def set_linear_config(self, option: str = None, z_value: str = None) -> None:
        """
        Setting simulation time parameters

        Parameters
        ----------
        option :
            examples for current implementations: linear_optimization_efficiency_costs,
             linear_optimization_calendaric_aging, linear_optimization_and_calendaric_aging
        z_value :
            Z-Value determines the maximum distance between the highest and lowest charged battery storage
            0.25 recommendet. Possible values [0 - 1].

        Returns
        -------

        """
        self._set(StorageSystemConfig.SECTION, StorageSystemConfig.LINEAR_OPTION, option)
        self._set(StorageSystemConfig.SECTION, StorageSystemConfig.LINEAR_DISTRIBUTION_Z_VALUE, z_value)

    def no_data_export(self) -> None:
        """
        No simulation results will be written to files

        Returns
        -------

        """
        self._set_bool(GeneralSimulationConfig.SECTION, GeneralSimulationConfig.EXPORT_DATA, False)

    def set_operation_strategy(self, strategy: str, min_soc: float = 0.0, max_soc: float = 1.0) -> None:
        """
        Setting the operation strategy

        Parameters
        ----------
        strategy :
            examples for current implementations: PowerFollower, SocFollower, ResidentialPvGreedy, ResidentialPvFeedInDamp, etc.
        min_soc :
            minimum allowed soc of storage technologies considered by the operation strategy
        max_soc :
            maximum allowed soc of storage technologies considered by the operation strategy

        Returns
        -------

        """
        self._set(EnergyManagementConfig.SECTION, EnergyManagementConfig.STRATEGY, strategy)
        self._set(EnergyManagementConfig.SECTION, EnergyManagementConfig.MIN_SOC, str(min_soc))
        self._set(EnergyManagementConfig.SECTION, EnergyManagementConfig.MAX_SOC, str(max_soc))

    def set_fcr_operation_strategy(self, fcr_power: float, idm_power: float, fcr_reserve: float = 0.25,
                                   soc_set: float = 0.52) -> None:
        """
        Setting Frequency Containment Reserve (FCR) including a Intraday Recharge Strategy (IDM) as the operation strategy

        Parameters
        ----------
        fcr_power :
            power to reserve for FCR
        idm_power :
            power to participate in IDM
        fcr_reserve :
            defining the lower and upper bounds for the energy capacity as an equivalent for time with full power in h
        soc_set :
            target value of the SOC considering system losses

        Returns
        -------

        """
        self.set_operation_strategy('FcrIdmRechargeStacked')
        self._set(EnergyManagementConfig.SECTION, EnergyManagementConfig.POWER_FCR, str(fcr_power))
        self._set(EnergyManagementConfig.SECTION, EnergyManagementConfig.POWER_IDM, str(idm_power))
        self._set(EnergyManagementConfig.SECTION, EnergyManagementConfig.FCR_RESERVE, str(fcr_reserve))
        self._set(EnergyManagementConfig.SECTION, EnergyManagementConfig.SOC_SET, str(soc_set))

    def set_ev_operation_strategy(self, operation_strategy: str, charging_strategy: str, max_power: float) -> None:
        """
        Setting Frequency Containment Reserve (FCR) including a Intraday Recharge Strategy (IDM) as the operation strategy

        Parameters
        ----------
        charging_strategy :
            select charging strategy (uncontrolled, ...)
        max_power :
            Maximum power to recharge when plugged in

        Returns
        -------

        """
        self.set_operation_strategy(operation_strategy)
        self._set(EnergyManagementConfig.SECTION, EnergyManagementConfig.EV_CHARGING_STRATEGY, charging_strategy)
        self._set(EnergyManagementConfig.SECTION, EnergyManagementConfig.MAX_POWER, str(max_power))

    def set_peak_shaving_strategy(self, strategy: str, max_power: float):
        """
        Setting peak shaving as the operation strategy

        Parameters
        ----------
        strategy :
            examples for current implementations: SimplePeakShaving, PeakShavingPerfectForesight, etc.
        max_power :
            maximum allowed profile power, operations strategy tries to reduce peak power to this value

        Returns
        -------

        """
        self._set(EnergyManagementConfig.SECTION, EnergyManagementConfig.STRATEGY, strategy)
        self._set(EnergyManagementConfig.SECTION, EnergyManagementConfig.MAX_POWER, str(max_power))

    def add_storage_system_ac(self, ac_power: float, intermediate_circuit_voltage: float,
                              acdc_converter: str, housing: str, hvac: str) -> str:
        """
        Adding an AC storage system to the config with the given parameters. All configured system will be instantiated.

        Parameters
        ----------
        ac_power :
            maximum AC power of storage system in W
        intermediate_circuit_voltage :
            voltage of the intermediate circuit in V
        acdc_converter :
            key from generated options for ACDC converter
        housing :
            key from generated options for housing
        hvac :
            key from generated options for hvac

        Returns
        -------
        str:
            key for AC storage system
        """
        key: str = self._get_id_from(StorageSystemConfig.SECTION, StorageSystemConfig.STORAGE_SYSTEM_AC)
        name: str = self.__AC_SYSTEM + key
        value: str = name + ','
        value += str(ac_power) + ','
        value += str(intermediate_circuit_voltage) + ','
        value += acdc_converter + ','
        value += housing + ','
        value += hvac
        self._add(StorageSystemConfig.SECTION, StorageSystemConfig.STORAGE_SYSTEM_AC, value)
        return name

    def clear_storage_system_ac(self) -> None:
        """
        Deleting all configured AC storage systems

        Returns
        -------

        """
        self._clear(StorageSystemConfig.SECTION, StorageSystemConfig.STORAGE_SYSTEM_AC)

    def add_storage_system_dc(self, ac_system_name: str, dcdc_converter: str, storage_name: str) -> None:
        """
        Adding an DC storage system to the config with the given parameters. Every DC systems needs to be connected to
        an AC storage system (via the given key). All configured system will be instantiated.

        Parameters
        ----------
        ac_system_name :
            key from generated options for AC storage systems
        dcdc_converter :
            key from generated options for DCDC converter
        storage_name :
            key from generated options for storage technologies

        Returns
        -------

        """
        value: str = ac_system_name + ','
        value += dcdc_converter + ','
        value += storage_name
        self._add(StorageSystemConfig.SECTION, StorageSystemConfig.STORAGE_SYSTEM_DC, value)

    def clear_storage_system_dc(self) -> None:
        """
        Deleting all configured DC storage systems

        Returns
        -------

        """
        self._clear(StorageSystemConfig.SECTION, StorageSystemConfig.STORAGE_SYSTEM_DC)

    def __add_storage_technology(self, capacity: float, storage_type: str, storage_characteristics: str) -> str:
        key: str = self._get_id_from(StorageSystemConfig.SECTION, StorageSystemConfig.STORAGE_TECHNOLOGY)
        name: str = self.__TECHNOLOGY + key
        value: str = name + ','
        value += str(capacity) + ','
        value += storage_type + ','
        value += storage_characteristics
        self._add(StorageSystemConfig.SECTION, StorageSystemConfig.STORAGE_TECHNOLOGY, value)
        return name

    def add_lithium_ion_battery(self, capacity: float, cell_type: str, start_soc: float = 0.5,
                                start_soh: float = 1.0) -> str:
        """
        Adding a lithium-ion battery to storage technology options with given parameters.

        Parameters
        ----------
        capacity :
            capacity of battery in Wh
        cell_type :
            examples for possible cell types: GenericCell, SonyLFP, PanasonicNCA, MolicelNMC, SanyoNMC
        start_soc :
            state of charge at start of simulation in p.u.
        start_soh :
            state of health at start of simulation in p.u. (Note: not all cell types are supported)

        Returns
        -------
        str:
            key for storage technology
        """
        value: str = cell_type + ','
        value += str(start_soc) + ','
        value += str(start_soh)
        return self.__add_storage_technology(capacity, 'lithium_ion', value)

    def add_generic_cell(self, capacity: float) -> str:
        """
        Convenience method for constructing a lithium-ion battery with a GenericCell

        Parameters
        ----------
        capacity :
            capacity of battery in Wh

        Returns
        -------
        str:
            key for storage technology
        """
        return self.add_lithium_ion_battery(capacity, 'GenericCell')

    def add_redox_flow_battery(self, capacity: float, stack_type: str, stack_power: float,
                               pump_algorithm: str = 'StoichFlowRate') -> str:
        """
        Adding a redox flow battery to storage technology options with given parameters.

        Parameters
        ----------
        capacity :
            capacity of battery in Wh
        stack_type :
            examples for possible stack types: CellDataStack5500W, DummyStack3000W, IndustrialStack1500W, etc.
        stack_power :
            maximum power of stack in W
        pump_algorithm :
            control algorithm for selected pump, default: StoichFlowRate

        Returns
        -------
        str:
            key for storage technology
        """
        value: str = stack_type + ','
        value += str(stack_power) + ','
        value += pump_algorithm
        return self.__add_storage_technology(capacity, 'redox_flow', value)

    def add_hydrogen_technology(self, capacity: float, fuel_cell: str, fuel_cell_power: float, electrolyzer: str,
                                electrolyzer_power: float, storage_type: str, pressure: float) -> str:
        """
        Adding a hydrogen energy chain to storage technology options with given parameters.

        Parameters
        ----------
        capacity :
            capacity of battery in Wh
        fuel_cell :
            examples for possible fuel cells: PemFuelCell, JupiterFuelCell
        fuel_cell_power :
            maximum power for fuel cell in W
        electrolyzer :
            examples for possible electrolyzers: PemElectrolyzer, AlkalineElectrolyzer, etc.
        electrolyzer_power :
            maximum power for electrolyzer in W
        storage_type :
            examples for possible storage types: PressureTank, SimplePipeline
        pressure :
            pressure for storage type in bar

        Returns
        -------
        str:
            key for storage technology
        """
        value: str = fuel_cell + ','
        value += str(fuel_cell_power) + ','
        value += electrolyzer + ','
        value += str(electrolyzer_power) + ','
        value += storage_type + ','
        value += str(pressure)
        return self.__add_storage_technology(capacity, 'hydrogen', value)

    def clear_storage_technology(self) -> None:
        """
        Deleting all configured storage technology options

        Returns
        -------

        """
        self._clear(StorageSystemConfig.SECTION, StorageSystemConfig.STORAGE_TECHNOLOGY)

    def add_dcdc_converter(self, converter_type: str, max_power: float, efficiency: float = None) -> str:
        """
        Adding a dcdc converter option to config

        Parameters
        ----------
        efficiency :
            efficiency of converter in p.u., only used for FixEfficiencyDcDcConverter
        converter_type :
            examples for DCDC converters: NoLossDcDcConverter, FixEfficiencyDcDcConverter, etc.
        max_power :
            maximum power of DCDC converter in W

        Returns
        -------
        str:
            key for dcdc converter
        """
        key: str = self._get_id_from(StorageSystemConfig.SECTION, StorageSystemConfig.DCDC_CONVERTER)
        name: str = self.__DCDC_CONVERTER + key
        value: str = name + ','
        value += converter_type + ','
        value += str(max_power)
        if efficiency:
            value += ',' + str(efficiency)
        self._add(StorageSystemConfig.SECTION, StorageSystemConfig.DCDC_CONVERTER, value)
        return name

    def add_no_loss_dcdc(self) -> str:
        """
        Convenience method to add a config option of a no loss converter

        Returns
        -------
        str:
            key for dcdc converter

        """
        return self.add_dcdc_converter('NoLossDcDcConverter', 0.0, 1.0)

    def add_fix_efficiency_dcdc(self, efficiency: float = 0.98) -> str:
        """
        Convenience method to add a config option of a fix efficiency converter

        Parameters
        ----------
        efficiency :
            efficiency of dcdc converter in p.u.

        Returns
        -------
        str:
            key for dcdc converter

        """
        return self.add_dcdc_converter('FixEfficiencyDcDcConverter', 0.0, efficiency)

    def clear_dcdc_converter(self) -> None:
        """
        Deleting all config options for DCDC converter

        Returns
        -------

        """
        self._clear(StorageSystemConfig.SECTION, StorageSystemConfig.DCDC_CONVERTER)

    def add_acdc_converter(self, converter_type: str, number_of_converters: int = 1, switch_value: float = 1.0) -> str:
        """
        Adding an acdc converter option to config

        Parameters
        ----------
        converter_type :
            examples for ACDC converters: NoLossAcDcConverter, FixEfficiencyAcDcConverter, BonfiglioliAcDcConverter, etc.
        number_of_converters :
            possibility to cascade converters with the given number, default: 1
        switch_value :
            if converters are cascaded, the switch value in p.u. defines the point of the power to nominal power ratio
            when the next converter will be activated, default: 1.0

        Returns
        -------
        str:
            key for acdc converter
        """
        key: str = self._get_id_from(StorageSystemConfig.SECTION, StorageSystemConfig.ACDC_CONVERTER)
        name: str = self.__ACDC_CONVERTER + key
        value: str = name + ','
        value += converter_type + ','
        value += str(number_of_converters) + ','
        value += str(switch_value)
        self._add(StorageSystemConfig.SECTION, StorageSystemConfig.ACDC_CONVERTER, value)
        return name

    def add_no_loss_acdc(self) -> str:
        """
        Convenience method to add a config option of a no loss converter

        Returns
        -------
        str:
            key for acdc converter
        """
        return self.add_acdc_converter('NoLossAcDcConverter')

    def add_fix_efficiency_acdc(self) -> str:
        """
        Convenience method to add a config option of a fix efficiency converter

        Returns
        -------
        str:
            key for acdc converter
        """
        return self.add_acdc_converter('FixEfficiencyAcDcConverter')

    def clear_acdc_converter(self) -> None:
        """
        Deleting all config options for ACDC converter

        Returns
        -------

        """
        self._clear(StorageSystemConfig.SECTION, StorageSystemConfig.ACDC_CONVERTER)

    def add_housing(self, housing_type: str, high_cube: bool = False, azimuth: float = 0.0,
                    absorptivity: float = 0.15, ground_albedo: float = 0.2) -> str:
        """
        Adding an housing option to config

        Parameters
        ----------
        housing_type :
            examples for housing: NoHousing, TwentyFtContainer, etc.
        high_cube :
            high cube container are taller than usual containers, default: False
        azimuth :
            azimuth angle
        absorptivity :
            absorptivity of container
        ground_albedo :
            reflection value of ground

        Returns
        -------
        str:
            key for housing
        """
        key: str = self._get_id_from(StorageSystemConfig.SECTION, StorageSystemConfig.HOUSING)
        name: str = self.__HOUSING + key
        value: str = name + ','
        value += housing_type + ','
        value += str(high_cube) + ','
        value += str(azimuth) + ','
        value += str(absorptivity) + ','
        value += str(ground_albedo)
        self._add(StorageSystemConfig.SECTION, StorageSystemConfig.HOUSING, value)
        return name

    def add_no_housing(self) -> str:
        """
        Convenience method to add a config option of a no housing

        Returns
        -------
        str:
            key for housing
        """
        return self.add_housing('NoHousing')

    def add_twenty_foot_container(self) -> str:
        """
        Convenience method to add a config option of a twenty foot container

        Returns
        -------
        str:
            key for housing
        """
        return self.add_housing('TwentyFtContainer')

    def clear_housing(self) -> None:
        """
        Deleting all config options for housing

        Returns
        -------

        """
        self._clear(StorageSystemConfig.SECTION, StorageSystemConfig.HOUSING)

    def add_hvac(self, hvac_type: str, power: float, temperature: float = 25.0) -> str:
        """
        Adding an HVAC option to config

        Parameters
        ----------
        hvac_type :
            examples for HVAC: NoHeatingVentilationAirConditioning, FixCOPHeatingVentilationAirConditioning, etc.
        power :
            maximum electrical heating/cooling power in W
        temperature :
            set point temperature in centigrade, default: 25.0

        Returns
        -------
        str:
            key for hvac system
        """
        key: str = self._get_id_from(StorageSystemConfig.SECTION, StorageSystemConfig.HVAC)
        name: str = self.__HVAC + key
        value: str = name + ','
        value += hvac_type + ','
        value += str(power) + ','
        value += str(temperature)
        self._add(StorageSystemConfig.SECTION, StorageSystemConfig.HVAC, value)
        return name

    def add_no_hvac(self) -> str:
        """
        Convenience method to add a config option of a no HVAC system

        Returns
        -------
        str:
            key for hvac system
        """
        return self.add_hvac('NoHeatingVentilationAirConditioning', 0.0)

    def add_constant_hvac(self, power: float, temperature: float) -> str:
        """
        Convenience method to add a config option of a constant HVAC system

        Parameters
        ----------
        power :
            maximum electrical heating/cooling power in W
        temperature :
            set point temperature in centigrade

        Returns
        -------
        str:
            key for hvac system
        """
        return self.add_hvac('FixCOPHeatingVentilationAirConditioning', power, temperature)

    def clear_hvac(self) -> None:
        """
        Deleting all config options for HVAC systems

        Returns
        -------

        """
        self._clear(StorageSystemConfig.SECTION, StorageSystemConfig.HVAC)

    def set_power_distribution_strategy_ac(self, strategy: str) -> None:
        """
        Defines the power distribution strategy for AC storage systems

        Parameters
        ----------
        strategy :
            examples: EqualPowerDistributor, SocBasedPowerDistributor, etc.

        Returns
        -------

        """
        self._set(StorageSystemConfig.SECTION, StorageSystemConfig.POWER_DISTRIBUTOR_AC, strategy)

    def set_power_distribution_strategy_dc(self, strategy: str) -> None:
        """
        Defines the power distribution strategy for DC storage systems

        Parameters
        ----------
        strategy :
            examples: EqualPowerDistributor, SocBasedPowerDistributor, etc.

        Returns
        -------

        """
        self._set(StorageSystemConfig.SECTION, StorageSystemConfig.POWER_DISTRIBUTOR_DC, strategy)

    def set_battery(self, start_soc: str = None, min_soc: str = None, max_soc: str = None,
                    eol: str = None, start_soh: str = None, exact_size: str = 'False') -> None:
        """
        Sets parameters for the battery section of the simulation.ini file
        :param start_soc: SOC at start of simulation
        :param min_soc: minimum permissible SOC
        :param max_soc: maximum permissible SOC
        :param eol: value of SOH at End-of-Life
        :param start_soh: SOH at start of simulation
        :param exact_size: Enable or disable rounding of number of cells in parallel/series to reach integer values
        :return:
        """
        self._set(BatteryConfig.SECTION, BatteryConfig.START_SOC, start_soc)
        self._set(BatteryConfig.SECTION, BatteryConfig.MIN_SOC, min_soc)
        self._set(BatteryConfig.SECTION, BatteryConfig.MAX_SOC, max_soc)
        self._set(BatteryConfig.SECTION, BatteryConfig.EOL, eol)
        self._set(BatteryConfig.SECTION, BatteryConfig.START_SOH, start_soh)
        self._set(BatteryConfig.SECTION, BatteryConfig.EXACT_SIZE, exact_size)

    def set_ambient_temperature_model(self, model: str, constant_temperature: float = None) -> None:
        """
        Set the type of ambient temperature model to be used
        :param constant_temperature: optional parameter for ConstantAmbientTemperature
        :param model: name of the ambient temperature model
        :return:
        """
        value = model
        if constant_temperature:
            value += ',' + str(constant_temperature)
        self._set(StorageSystemConfig.SECTION, StorageSystemConfig.AMBIENT_TEMPERATURE_MODEL, value)

    def set_ambient_temperature_profile_file(self, filename: str) -> None:
        """
        Set the filename for the location ambient temperature profile
        :param filename: name of specified file as str
        :return: None
        """
        self._set(ProfileConfig.SECTION, ProfileConfig.AMBIENT_TEMPERATURE_PROFILE, filename)

    def set_ghi_profile_file(self, filename: str) -> None:
        """
        Set the filename for the location Global Horizontal Irradiance profile
        :param filename: name of specified file as str
        :return: None
        """
        self._set(ProfileConfig.SECTION, ProfileConfig.GLOBAL_HORIZONTAL_IRRADIATION_PROFILE, filename)

    def set_solar_irradiation_model(self, model: str) -> None:
        """
        Set the type of solar irradiation model
        :param model: name of the ambient temperature model
        :return: None
        """
        self._set(StorageSystemConfig.SECTION, StorageSystemConfig.SOLAR_IRRADIATION_MODEL, model)

    def set_enable_thermal_simulation(self, value: str) -> None:
        """
        Enable/disable thermal simulation
        :param value: True/False as str
        :return: None
        """
        self._set(StorageSystemConfig.SECTION, StorageSystemConfig.THERMAL_SIMULATION, value)

    def set_load_scaling_factor(self, load_scaling_factor: str) -> None:
        """
        Sets the load scaling factor for the area under the load profile (energy)
        Parameters
        ----------
        load_scaling_factor : desired load scaling factor as str
            examples: 5e6, 5e7, etc.

        Returns
        -------

        """
        self._set(ProfileConfig.SECTION, ProfileConfig.LOAD_SCALING_FACTOR, load_scaling_factor)

    def set_load_generation_scaling_factor(self, load_generation_scaling_factor: str) -> None:
        """
        Sets the generation load scaling factor for the area under the load profile (energy)
        Parameters
        ----------
        load_generation_scaling_factor : desired load scaling factor as str
            examples: 5e6, 5e7, etc.

        Returns
        -------

        """
        self._set(ProfileConfig.SECTION, ProfileConfig.GENERATION_SCALING_FACTOR, load_generation_scaling_factor)

    def set_profile_config(self, load_profile: str = None) -> None:
        """
        Set the filename for the load profile
        :param load_profile: name of specified file as str
        :return: None
        """
        self._set(ProfileConfig.SECTION, ProfileConfig.LOAD_PROFILE, load_profile)

    def set_profile_direction(self, all_profile_direction: str) -> None:
        """
        Sets the parameters for the load profile: Direction
        Parameters
        ----------

        Returns
        -------

        """
        self._set(ProfileConfig.SECTION, ProfileConfig.POWER_PROFILE_DIR, all_profile_direction)
        self._set(ProfileConfig.SECTION, ProfileConfig.TECHNICAL_PROFILE_DIR, all_profile_direction)


    def set_total_profile_config(self, all_profile_direction, soc_profile, binary_profile, v2gprofile: str = None) -> None:
        """
        Sets parameters for the profile section of the simulation.ini
        :param technical_profile_direction: name of the profile direction as str
        :param soc_profile: name of specified soc profile as str
        :param binary_profile: name of specified binary profile as str

        :return:
        """
        self._set(ProfileConfig.SECTION, ProfileConfig.TECHNICAL_PROFILE_DIR, all_profile_direction)
        self._set(ProfileConfig.SECTION, ProfileConfig.POWER_PROFILE_DIR, all_profile_direction)
        self._set(ProfileConfig.SECTION, ProfileConfig.SOC_PROFILE, soc_profile)
        self._set(ProfileConfig.SECTION, ProfileConfig.BINARY_PROFILE, binary_profile)
        self._set(ProfileConfig.SECTION, ProfileConfig.V2G_PROFILE, v2gprofile)

    def set_generation_profile_config(self, generation_profile: str = None) -> None:
        """
        Set the filename for the generation profile
        :param generation_profile: name of specified file as str
        :return: None
        """
        self._set(ProfileConfig.SECTION, ProfileConfig.GENERATION_PROFILE, generation_profile)

    def set_generation_profile_parameters(self, generation_profile: str, generation_scaling_type: str,
                                          generation_scaling_factor: str) -> None:
        """
       Sets the parameters for the generation profile: Direction, name of profile, generation scaling type (energy/power),
        generation scaling factor
        Parameters
        ----------
        load_generation_scaling_factor : desired load scaling factor as str
            examples: 5e6, 5e7, etc.

        Returns
        -------

        """
        self._set(ProfileConfig.SECTION, ProfileConfig.GENERATION_PROFILE, generation_profile)
        self._set(ProfileConfig.SECTION, ProfileConfig.GENERATION_PROFILE_SCALING, generation_scaling_type)
        self._set(ProfileConfig.SECTION, ProfileConfig.GENERATION_SCALING_FACTOR, generation_scaling_factor)

    def set_load_profile_parameters(self, all_profile_direction: str, load_profile: str, load_scaling_type: str,
                                    load_scaling_factor: str) -> None:
        """
        Sets the parameters for the load profile: Direction, name of profile, load scaling type (energy/power),
        load scaling factor
        Parameters
        ----------
        load_generation_scaling_factor : desired load scaling factor as str
            examples: 5e6, 5e7, etc.

        Returns
        -------

        """
        self._set(ProfileConfig.SECTION, ProfileConfig.POWER_PROFILE_DIR, all_profile_direction)
        self._set(ProfileConfig.SECTION, ProfileConfig.LOAD_PROFILE, load_profile)
        self._set(ProfileConfig.SECTION, ProfileConfig.LOAD_PROFILE_SCALING, load_scaling_type)
        self._set(ProfileConfig.SECTION, ProfileConfig.LOAD_SCALING_FACTOR, load_scaling_factor)


    def set_frequency_profile_config(self, frequency_profile: str = None) -> None:
        """
        Set the filename for the frequency profile
        :param frequency_profile: name of specified file as str
        :return: None
        """
        self._set(ProfileConfig.SECTION, ProfileConfig.FREQUENCY_PROFILE, frequency_profile)



from simses.commons.config.simulation.energy_management import EnergyManagementConfig
from simses.commons.config.simulation.general import GeneralSimulationConfig
from simses.commons.config.simulation.profile import ProfileConfig
from simses.logic.energy_management.strategy.basic.frequency_containment_reserve import \
    FrequencyContainmentReserve
from simses.logic.energy_management.strategy.basic.intraday_market_recharge import \
    IntradayMarketRecharge
from simses.logic.energy_management.strategy.operation_priority import OperationPriority
from simses.logic.energy_management.strategy.stacked.stacked_operation_strategy import \
    StackedOperationStrategy


class FcrIdmRechargeStacked(StackedOperationStrategy):

    def __init__(self, general_config: GeneralSimulationConfig, ems_config: EnergyManagementConfig,
                 profile_config: ProfileConfig):
        super().__init__(OperationPriority.VERY_HIGH, [FrequencyContainmentReserve(general_config, ems_config, profile_config),
                                                       IntradayMarketRecharge(general_config, ems_config)])



from simses.commons.log import Logger
from simses.commons.state.technology.redox_flow import RedoxFlowState
from simses.technology.redox_flow.pump_algorithm.abstract_pump_algorithm import PumpAlgorithm
from simses.technology.redox_flow.stack.electrolyte.abstract_electrolyte import ElectrolyteSystem
from simses.technology.redox_flow.stack.abstract_stack import StackModule


class ConstantPressurePulsed(PumpAlgorithm):
    """Pump algorithm with a fix pressure drop that pulses the discharge and charge flow rate for a certain time period
    when the voltage is drops below the minimal voltage (discharging) or rises above the maximal voltage (charging)."""

    """The value for the pressure drop needs to be specified for the selected stack module. This pump algorithm should 
    only be used when the selected stack type considers changes in the internal resistance due to the pump stops. This
    is currently implemented for the stack modules IndustrialStack1500W and IndustrialStack9000W."""
    __PRESSURE_DROP = 16000  # Pa
    __PUMP_TIME = 90  # s

    def __init__(self, stack_module: StackModule, pump, electrolyte_system: ElectrolyteSystem, time_step: int):
        super().__init__(pump)
        self.__log: Logger = Logger(type(self).__name__)
        self.__stack_module: StackModule = stack_module
        self.__electrolyte_system: ElectrolyteSystem = electrolyte_system
        self.__pressure_drop_anolyte = self.__PRESSURE_DROP
        self.__pressure_drop_catholyte = self.__PRESSURE_DROP
        self.__pump_time = self.__PUMP_TIME
        self.__hydraulic_resistance = stack_module.get_hydraulic_resistance()
        self.__max_voltage = self.__stack_module.get_max_voltage()
        self.__min_voltage = self.__stack_module.get_min_voltage()
        self.__time_step = time_step

    def get_pressure_drop_catholyte(self, redox_flow_state: RedoxFlowState):
        voltage = redox_flow_state.voltage
        power = redox_flow_state.power
        time_pump = redox_flow_state.time_pump

        if (((voltage >= self.__max_voltage or voltage <= self.__min_voltage) or 0 < time_pump < self.__pump_time) and
                power != 0):
            self.__pressure_drop_catholyte = self.__PRESSURE_DROP
            if time_pump < 0:
                time_pump = 0
            time_pump += self.__time_step
        else:
            self.__pressure_drop_catholyte = 0
            if time_pump > 0:
                time_pump = 0
            time_pump -= self.__time_step

        redox_flow_state.time_pump = time_pump
        return self.__pressure_drop_catholyte

    def get_pressure_drop_anolyte(self, redox_flow_state: RedoxFlowState):
        voltage = redox_flow_state.voltage
        power = redox_flow_state.power
        time_pump = redox_flow_state.time_pump

        if (((voltage > self.__max_voltage or voltage < self.__min_voltage) or 0 < time_pump < self.__pump_time)
                and power != 0):
            self.__pressure_drop_anolyte = self.__PRESSURE_DROP
        else:
            self.__pressure_drop_anolyte = 0

        return self.__pressure_drop_anolyte

    def get_flow_rate_anolyte(self, redox_flow_state: RedoxFlowState):
        flow_rate = (self.__pressure_drop_catholyte / (self.__hydraulic_resistance *
                     self.__electrolyte_system.get_viscosity_anolyte(redox_flow_state)))
        return flow_rate

    def get_flow_rate_catholyte(self, redox_flow_state: RedoxFlowState):
        flow_rate = (self.__pressure_drop_catholyte / (self.__hydraulic_resistance *
                     self.__electrolyte_system.get_viscosity_catholyte(redox_flow_state)))
        return flow_rate

    def get_flow_rate_max(self):
        flow_rate_max = (self.__PRESSURE_DROP / (self.__hydraulic_resistance *
                         self.__electrolyte_system.get_min_viscosity()))
        return flow_rate_max

    def get_flow_rate_min(self):
        flow_rate_min = (self.__PRESSURE_DROP / (self.__hydraulic_resistance *
                         self.__electrolyte_system.get_max_viscosity()))
        return flow_rate_min

    def close(self):
        super().close()
        self.__log.close()
        self.__stack_module.close()
        self.__electrolyte_system.close()



from simses.commons.state.technology.storage import StorageTechnologyState


class FuelCellState(StorageTechnologyState):
    """
    Current physical state of the fuel cell components with the main electrical parameters.
    """

    SYSTEM_AC_ID = 'StorageSystemAC'
    SYSTEM_DC_ID = 'StorageSystemDC'
    VOLTAGE = 'voltage of fuel cell in V'
    CURRENT = 'current of fuel cell in A'
    CURRENT_DENSITY = 'current density of fuel cell in A cm-2'
    POWER = 'power in W'
    TEMPERATURE = 'temperature of fuel cell stack in K'
    POWER_LOSS = 'Power loss in W'
    HYDROGEN_USE = 'hydrogen use in mol/s'  # hydrogen use of fuel cell
    HYDROGEN_INFLOW = 'hydrogen inflow in mol/s'  # hydrogen inflow into fuelcell stack for pressure control
    OXYGEN_USE = 'oxygen use in mol/s'  # oxygen use of fuel cell
    OXYGEN_INFLOW = 'oxygen inflow in mol/s'  # oxygen inflow into fuelcell stack for pressure control of fuel cell
    FULFILLMENT = 'fulfillment in p.u.'  # ??
    CONVECTION_HEAT = 'convection heat in W'
    PRESSURE_ANODE = 'relative anode pressure of fuelcell in barg'
    PRESSURE_CATHODE = 'relative cathode pressure of fuelcell in barg'
    SOH = 'State of health in p.u.'
    SOC = 'SOC of Hydrogen Storage'
    MAX_CHARGE_POWER = 'max charge power in W'
    MAX_DISCHARGE_POWER = 'max discharge power in W'

    def __init__(self, system_id: int, storage_id: int):
        super().__init__()
        self._initialize()
        self.set(self.SYSTEM_AC_ID, system_id)
        self.set(self.SYSTEM_DC_ID, storage_id)

    @property
    def soc(self) -> float:
        pass

    @soc.setter
    def soc(self, value: float) -> None:
        pass

    @property
    def capacity(self) -> float:
        pass

    @capacity.setter
    def capacity(self, value: float):
        pass

    @property
    def voltage(self) -> float:
        return self.get(self.VOLTAGE)

    @voltage.setter
    def voltage(self, value: float) -> None:
        self.set(self.VOLTAGE, value)

    @property
    def current(self) -> float:
        return self.get(self.CURRENT)

    @current.setter
    def current(self, value: float) -> None:
        self.set(self.CURRENT, value)

    @property
    def current_density(self) -> float:
        return self.get(self.CURRENT_DENSITY)

    @current_density.setter
    def current_density(self, value: float) -> None:
        self.set(self.CURRENT_DENSITY, value)

    @property
    def temperature(self) -> float:
        return self.get(self.TEMPERATURE)

    @temperature.setter
    def temperature(self, value: float) -> None:
        self.set(self.TEMPERATURE, value)

    @property
    def power_loss(self) -> float:
        return self.get(self.POWER_LOSS)

    @power_loss.setter
    def power_loss(self, value: float) -> None:
        self.set(self.POWER_LOSS, value)

    # @property
    # def power(self) -> float:
    #     return self.get(self.POWER)
    #
    # @power.setter
    # def power(self, value: float) -> None:
    #     self.set(self.POWER, value)

    @property
    def hydrogen_use(self) -> float:
        return self.get(self.HYDROGEN_USE)

    @hydrogen_use.setter
    def hydrogen_use(self, value: float) -> None:
        self.set(self.HYDROGEN_USE, value)

    @property
    def hydrogen_inflow(self) -> float:
        return self.get(self.HYDROGEN_INFLOW)

    @hydrogen_inflow.setter
    def hydrogen_inflow(self, value: float) -> None:
        self.set(self.HYDROGEN_INFLOW, value)

    @property
    def oxygen_use(self) -> float:
        return self.get(self.OXYGEN_USE)

    @oxygen_use.setter
    def oxygen_use(self, value: float) -> None:
        self.set(self.OXYGEN_USE, value)

    @property
    def oxygen_inflow(self) -> float:
        return self.get(self.OXYGEN_INFLOW)

    @oxygen_inflow.setter
    def oxygen_inflow(self, value: float) -> None:
        self.set(self.OXYGEN_INFLOW, value)

    @property
    def convection_heat(self) -> float:
        return self.get(self.CONVECTION_HEAT)

    @convection_heat.setter
    def convection_heat(self, value: float) -> None:
        self.set(self.CONVECTION_HEAT, value)

    @property
    def fulfillment(self) -> float:
        return self.get(self.FULFILLMENT)

    @fulfillment.setter
    def fulfillment(self, value: float):
        self.set(self.FULFILLMENT, value)

    @property
    def pressure_anode(self) -> float:
        return self.get(self.PRESSURE_ANODE)

    @pressure_anode.setter
    def pressure_anode(self, value: float):
        self.set(self.PRESSURE_ANODE, value)

    @property
    def pressure_cathode(self) -> float:
        return self.get(self.PRESSURE_CATHODE)

    @pressure_cathode.setter
    def pressure_cathode(self, value: float):
        self.set(self.PRESSURE_CATHODE, value)

    @property
    def soh(self) -> float:
        return self.get(self.SOH)

    @soh.setter
    def soh(self, value: float):
        self.set(self.SOH, value)

    @property
    def max_charge_power(self) -> float:
        return self.get(self.MAX_CHARGE_POWER)

    @max_charge_power.setter
    def max_charge_power(self, value: float):
        self.set(self.MAX_CHARGE_POWER, value)

    @property
    def max_discharge_power(self) -> float:
        return self.get(self.MAX_DISCHARGE_POWER)

    @max_discharge_power.setter
    def max_discharge_power(self, value: float):
        self.set(self.MAX_DISCHARGE_POWER, value)

    @property
    def id(self) -> str:
        return 'FUELCELL' + str(self.get(self.SYSTEM_AC_ID)) + str(self.get(self.SYSTEM_DC_ID))

    @property
    def is_charge(self) -> bool:
        return self.current > 0

    @classmethod
    def sum_parallel(cls, hydrogen_states: []):
        hydrogen_state = FuelCellState(0, 0)
        return hydrogen_state

    @classmethod
    def sum_serial(cls, states: []):
        pass


from simses.commons.config.simulation.battery import BatteryConfig
from simses.commons.cycle_detection.cycle_detector import CycleDetector
from simses.technology.lithium_ion.cell.type import CellType
from simses.technology.lithium_ion.degradation.calendar.nmc_sanyo_ur18650e import \
    SanyoNMCCalendarDegradationModel
from simses.technology.lithium_ion.degradation.cyclic.nmc_sanyo_ur18650e import \
    SanyoNMCCyclicDegradationModel
from simses.technology.lithium_ion.degradation.degradation_model import DegradationModel


class SanyoNMCDegradationModel(DegradationModel):
    def __init__(self, cell_type: CellType, cycle_detector: CycleDetector, battery_config: BatteryConfig):
        super().__init__(cell_type, SanyoNMCCyclicDegradationModel(cell_type, cycle_detector),
                         SanyoNMCCalendarDegradationModel(cell_type), cycle_detector,
                         battery_config)


from simses.commons.state.abstract_state import State


class EnergyManagementState(State):
    """
    Current State of the Energy Management (PV, Load, etc..)
    """

    AC_POWER_REQUESTED = 'Power requested from EMS in W'
    LOAD_POWER = 'Load in W'
    PV_POWER = 'PV Generation in W'
    FCR_MAX_POWER = 'Power reserved for FCR in W'
    IDM_POWER = 'Power delivered for IDM in W'
    PEAKSHAVING_LIMIT = 'Peak Shaving Limit in W'
    BINARY_VALUE = 'Binary value for EVs (1 available, 0 on the road)'
    V2G_POWER = 'Power provided in V2G operation'

    def __init__(self):
        super().__init__()
        self._initialize()

    @property
    def ac_power_requested(self) -> float:
        return self.get(self.AC_POWER_REQUESTED)

    @ac_power_requested.setter
    def ac_power_requested(self, value: float) -> None:
        self.set(self.AC_POWER_REQUESTED, value)

    @property
    def load_power(self) -> float:
        return self.get(self.LOAD_POWER)

    @load_power.setter
    def load_power(self, value: float) -> None:
        self.set(self.LOAD_POWER, value)

    @property
    def pv_power(self) -> float:
        return self.get(self.PV_POWER)

    @pv_power.setter
    def pv_power(self, value: float) -> None:
        self.set(self.PV_POWER, value)

    @property
    def binary(self) -> float:
        return self.get(self.BINARY_VALUE)

    @binary.setter
    def binary(self, value: float) -> None:
        self.set(self.BINARY_VALUE, value)

    @property
    def v2g_power(self) -> float:
        return self.get(self.V2G_POWER)

    @v2g_power.setter
    def v2g_power(self, value: float) -> None:
        self.set(self.V2G_POWER, value)

    @property
    def fcr_max_power(self) -> float:
        return self.get(self.FCR_MAX_POWER)

    @fcr_max_power.setter
    def fcr_max_power(self, value: float) -> None:
        self.set(self.FCR_MAX_POWER, value)

    @property
    def idm_power(self) -> float:
        return self.get(self.IDM_POWER)

    @idm_power.setter
    def idm_power(self, value: float) -> None:
        self.set(self.IDM_POWER, value)

    @property
    def peakshaving_limit(self) -> float:
        return self.get(self.PEAKSHAVING_LIMIT)

    @peakshaving_limit.setter
    def peakshaving_limit(self, value: float) -> None:
        self.set(self.PEAKSHAVING_LIMIT, value)

    @property
    def id(self) -> str:
        return 'EMS'

    @classmethod
    def sum_parallel(cls, system_states: []):
        raise Exception('sum_parallel is not implemented for EnergyManagementState')

    @classmethod
    def sum_serial(cls, states: []):
        raise Exception('sum_serial is not implemented for EnergyManagementState')



class Layer:

    """
    Class Layer specifies and controls the attributes of each layer of the wall of a shipping container (Housing)
    """

    # Layer attribute names
    LENGTH: str = 'length'
    BREADTH: str = 'breadth'
    HEIGHT: str = 'height'
    THICKNESS: str = 'thickness'
    DENSITY: str = 'density'
    THERMAL_CONDUCTIVITY: str = 'thermal_conductivity'
    SPECIFIC_HEAT: str = 'specific_heat'
    CONVECTION_COEFFICIENT: str = 'convection_coefficient_air'
    ABSORPTIVITY: str = 'absorptivity'
    TEMPERATURE: str = 'temperature'

    def __init__(self, layer_attributes: dict):
        self.__length = layer_attributes[self.LENGTH]
        self.__breadth = layer_attributes[self.BREADTH]
        self.__height = layer_attributes[self.HEIGHT]
        self.__thickness = layer_attributes[self.THICKNESS]
        self.__density = layer_attributes[self.DENSITY]
        self.__thermal_conductivity = layer_attributes[self.THERMAL_CONDUCTIVITY]
        self.__specific_heat = layer_attributes[self.SPECIFIC_HEAT]
        self.__convection_coefficient_air = layer_attributes[self.CONVECTION_COEFFICIENT]
        self.__absorptivity = layer_attributes[self.ABSORPTIVITY]
        self.__temperature = layer_attributes[self.TEMPERATURE]

        self.__surface_area_long_side = self.__length * self.__height
        self.__surface_area_roof = self.__length * self.__breadth
        self.__surface_area_short_side = self.__breadth * self.__height
        self.__surface_area_total = 2 * (self.__surface_area_long_side + self.__surface_area_short_side + self.__surface_area_roof)
        self.__material_volume = self.__surface_area_total * self.__thickness
        self.__mass = self.__material_volume * self.__density
        self.__scale_factor = 1

    def update_scale(self):
        self.__scale_factor += 1

    @property
    def length(self) -> float:
        return self.__length

    @property
    def breadth(self) -> float:
        return self.__breadth

    @property
    def height(self) -> float:
        return self.__height

    @property
    def thickness(self) -> float:
        return self.__thickness

    @property
    def mass(self) -> float:
        return self.__mass

    @property
    def thermal_conductivity(self) -> float:
        return self.__thermal_conductivity

    @property
    def specific_heat(self) -> float:
        return self.__specific_heat

    @property
    def convection_coefficient_air(self) -> float:
        return self.__convection_coefficient_air

    @property
    def absorptivity(self) -> float:
        return self.__absorptivity

    @property
    def surface_area_long_side(self) -> float:
        return self.__surface_area_long_side * self.__scale_factor

    @property
    def surface_area_short_side(self) -> float:
        return self.__surface_area_short_side * self.__scale_factor

    @property
    def surface_area_roof(self) -> float:
        return self.__surface_area_roof * self.__scale_factor

    @property
    def surface_area_total(self) -> float:
        return self.__surface_area_total * self.__scale_factor

    @property
    def temperature(self) -> float:
        return self.__temperature



from simses.commons.config.simulation.battery import BatteryConfig
from simses.commons.cycle_detection.cycle_detector import CycleDetector
from simses.technology.lithium_ion.cell.type import CellType
from simses.technology.lithium_ion.degradation.calendar.no_calendar_dedradation import \
    NoCalendarDegradationModel
from simses.technology.lithium_ion.degradation.cyclic.no_cyclic_degradation import \
    NoCyclicDegradationModel
from simses.technology.lithium_ion.degradation.degradation_model import DegradationModel


class NoDegradationModel(DegradationModel):

    def __init__(self, cell: CellType, cycle_detector:CycleDetector, battery_config: BatteryConfig):
        super().__init__(cell, NoCyclicDegradationModel(), NoCalendarDegradationModel(), cycle_detector, battery_config)


import math as m
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt


TEMP_IDX = 0
P_A_IDX = 1
P_C_IDX = 2
POWER_DENS_IDX = 3
CURRENT_DENS_IDX = 4
ALPHA_IDX = 5
I_0_IDX = 6

# constants
IDEAL_GAS_CONST: float = 8.314462  # J/(mol K)
FARADAY_CONST: float = 96485.3321  # As/mol
A_CATHODE = 2.4  # bar cm²/A  -> angaben, Abhänigkeit von Dicke
A_ANODE = 2.8  # bar cm²/A
R_ele = 0.096  # Ohm cm²
LAMBDA = 25  # degree of humidification
THIKNESS_NAFION = 200  # um

# import of look up table
current_dens_lookup_data = pd.read_csv('cell_currentdensity_lookup_with_alpha.csv')
#current_dens_lookup_data = pd.read_csv('cell_currentdensity_lookup.csv')
temperature = current_dens_lookup_data.iloc[:, TEMP_IDX]
p_anode = current_dens_lookup_data.iloc[:, P_A_IDX]
p_cathode = current_dens_lookup_data.iloc[:, P_C_IDX]
power_dens = current_dens_lookup_data.iloc[:, POWER_DENS_IDX]
current_dens = current_dens_lookup_data.iloc[:, CURRENT_DENS_IDX]
alpha = current_dens_lookup_data.iloc[:, ALPHA_IDX]
i_0 = current_dens_lookup_data.iloc[:, I_0_IDX]

# parameters of curve fitting of activation powerdensity
p00 = 0
p10 = -0.007005
p01 = 0
p20 = 0.01346
p11 = -0.0001083
p02 = 0

current_dens_analytic = []

for i in range(len(temperature)):
    U_rev = 1.5184 - 1.5421 * 10 ** (-3) * (temperature[i] + 273.15) + 9.523 * 10 ** (-5) * (temperature[i]  + 273.15) * m.log(temperature[i]  + 273.15) \
                         + 9.84 * 10 ** (-8) * (temperature[i]  + 273.15) ** 2

    conductivity_nafion = (0.005139 * LAMBDA - 0.00326) * m.exp(1268 * (1 / 303 - 1 / (temperature[i]  + 273.15)))  # S/cm 0.14
    R_mem = THIKNESS_NAFION * 10 ** (-4) / conductivity_nafion  # Ohm cm2

    # calculate partial pressures of hydrogen and oxygen for calculation of Nernst-equation
    pressure_sat_h2o = 10 ** (-2.1794 + 0.02953 * temperature[i]  - 9.1837 * 10 ** (-5) * temperature[i]  ** 2 + 1.4454 * 10 ** (-7) * temperature[i] ** 3)
    p_h2 = (1 + p_cathode[i]) - pressure_sat_h2o  # without currentdensity dependency
    p_o2 = (1 + p_anode[i]) - pressure_sat_h2o  # without currentdensity dependency
    p_h2_ref = 1  # bar
    p_o2_ref = 1  # bar

    a = R_mem + R_ele + p20
    b = U_rev + IDEAL_GAS_CONST * (temperature[i]  + 273.15) / (2 * FARADAY_CONST) * m.log((p_o2 / p_o2_ref)**(1/2) * p_h2 / p_h2_ref) + p10 + p11 * temperature[i] - alpha[i] * m.log(i_0[i])
    c = p00 + p01 * temperature[i]  + p02 * temperature[i]  ** 2 - power_dens[i]

    current_dens_analytic.append((- b + (b ** 2 - 4 * a * c) ** (1/2)) / (2 * a))
    #i2 = (- b - (b ** 2 - 4 * a * c) ** (1/2)) / (2 * a)

abs_err = current_dens - current_dens_analytic

mean_abs_err = np.mean(abs(abs_err))

max_abs_err = max(abs(abs_err))

RMS = np.sqrt(sum((abs_err)**2)/len(abs_err))

x = list(range(len(temperature)))
y = abs_err
plt.ylabel('absolut error')
plt.title('error analytic calculated current vs. look up table')
plt.plot(x, y, 'rx')
plt.figtext(0.15, 0.25, "mean abs. error: %.6f"%mean_abs_err, fontweight="bold")
plt.figtext(0.15, 0.2, "RMS error: %.6f"%RMS, fontweight="bold")
plt.figtext(0.15, 0.15, "maximal absolut error: %.6f"%max_abs_err, fontweight="bold")
plt.grid(True)
#plt.savefig('error_of_analytic_currentdensity_calculation')
plt.savefig('error_of_analytic_currentdensity_calculation')
plt.show()





from simses.commons.log import Logger
from simses.commons.state.parameters import SystemParameters
from simses.commons.state.technology.lithium_ion import LithiumIonState
from simses.commons.utils.utilities import get_average_from
from simses.system.auxiliary.auxiliary import Auxiliary
from simses.technology.lithium_ion.battery import LithiumIonBattery
from simses.technology.storage import StorageTechnology


class BatteryCircuit(StorageTechnology):
    """
    BatteryCircuit is tasked to directly connect several battery types. It handles the balancing currents between
    the batteries by equalizing the voltage difference.

    TODO

    1) Having static max / min values for current and voltage or dynamic for each battery type and state?

    -> Not needed anymore

    2) How to calculate the equilibrium balancing current between batteries?

    -> Calculate balancing current depending on OCV an internal resistence of each battery

    3) How to avoid divergency for specific LIB types? Are LIB types correctly modelled? Why do some work, others dont?
        Working: PanasonicNCA, MolicelNMC, SanyoNMC, AkasolAkmNMC;
        Not working: SonyLFP, GenericCell, AkasolOemNMC;

    -> All battery types should work

    4) Current alogrithm works better with higher capacities / lower C-Rates

    5) Right now the overall current is split between the batteries depending on their capacities. One way to a faster voltage convergence could be to only discharge the battery with higher voltage and charge only the battery with a lower voltage with the applied current.

    -> Current is split depending on OCV and internal resistance, in the same function which also calculate the balance currents

    6) During the balancing process batteries partly cannot handle the balancing current, hence it takes time for converging - in standard simulation around 30 min.

    -> If batteries can not handle the balancing current or overall current (BMS is limiting), the BMS would open the circuit breakers to protect the batterie from overcurrent.
       If this happens, the scenario is not feasable.

    """

    __LARGE_POSITIVE_NUMBER: float = 1e100
    __LARGE_NEGATIVE_NUMBER: float = -1e100

    __VOLTAGE_ACCURACY: float = 1.0
    """Acceptable voltage difference between batteries in V. Possibility of infinite loops if set too low."""

    __CURRENT_ACCURACY: float = 0.1
    """Minimum current applied to batteries in A."""

    __MAX_SLOPE: float = 1.0
    """Maximum change of current per cycle. Trade off between overshooting voltage against performance."""

    def __init__(self, batteries: [LithiumIonBattery]):
        super(BatteryCircuit, self).__init__()
        self.__log: Logger = Logger(type(self).__name__)
        self.__batteries: [LithiumIonBattery] = batteries
        self.__battery_dict: dict = dict()
        self.__dc_switch_operations: dict = dict()
        self.__last_active_battery_states: dict = dict()
        for battery in batteries:
            self.__battery_dict[battery.id] = battery
            self.__dc_switch_operations[battery.id] = 0
            self.__last_active_battery_states[battery.id] = battery.state
        max_voltage: float = self.__LARGE_POSITIVE_NUMBER
        #ref_min_voltage: float = self.__LARGE_POSITIVE_NUMBER
        #ref_min_voltage: float = 0.0
        min_voltage: float = 0.0
        max_current: float = self.__LARGE_POSITIVE_NUMBER
        min_current: float = self.__LARGE_NEGATIVE_NUMBER
        for battery in batteries:
            max_voltage = min(max_voltage, battery.max_voltage)
            min_voltage = max(min_voltage, battery.min_voltage)
            max_current = min(max_current, battery.max_current)
            min_current = max(min_current, battery.min_current)
        self.__max_voltage: float = max_voltage
        self.__min_voltage: float = min_voltage
        self.__max_current: float = min(max_current, abs(min_current)) * self.__MAX_SLOPE
        self.__switch_current: bool = True

    def __get_mean_voltage(self, battery_states: dict) -> float:
        voltages: [float] = list()
        for battery_state in battery_states.values():
            battery_state: LithiumIonState = battery_state
            voltages.append(battery_state.voltage)
        return get_average_from(voltages)

    def __get_battery_current(self, battery_states: dict, battery_id: int, current: float) -> float:
        """Get the current for one Battery in a direct parallel (before DC/DC) setup"""
        upper_sum: float = 0.0
        lower_sum: float = 0.0
        for battery in battery_states.keys():
            if battery != battery_id:
                if not (battery_states[battery].internal_resistance and battery_states[battery_id].internal_resistance):
                    self.__log.error('Internal resistance of at least one direct parallel battery is 0')
                upper_sum += (battery_states[battery].voltage_open_circuit - battery_states[
                    battery_id].voltage_open_circuit) / battery_states[battery].internal_resistance
                lower_sum += battery_states[battery_id].internal_resistance / battery_states[
                    battery].internal_resistance
        return (current + upper_sum) / (1 + lower_sum)

    @staticmethod
    def __sort_by_soc(data: dict, descending: bool = False) -> dict:
        return dict(sorted(data.items(), key=lambda x: x[1].soc, reverse=descending))

    @staticmethod
    def __sort_list_by_soc(data: list, descending: bool = False) -> list:
        return sorted(data, key=lambda x: x.state.soc, reverse=descending)

    def __balance_batteries(self, time: float, current: float) -> dict:
        #self.__batteries = self.__sort_list_by_voltage(self.__batteries, False)
        local_battery_states: dict = dict()
        active_local_battery_states: dict = dict()
        for battery in self.__batteries:
            local_battery_states[battery.id] = battery.state
            # active battery states: if a direct parallel battery gets empty it disconnects from DC bus temporarily, hence gets removed from active
            active_local_battery_states[battery.id] = battery.state
        min_voltage = self.__LARGE_POSITIVE_NUMBER
        max_voltage = 0.0

        active_local_battery_states, local_battery_states = self.__filter_battery_states_by_voltage(
            active_local_battery_states, current, local_battery_states)

        loop_count: int = 0
        while abs(max_voltage - min_voltage) > self.__VOLTAGE_ACCURACY:
            loop_count += 1
            min_voltage = self.__LARGE_POSITIVE_NUMBER
            max_voltage = 0
            # First calculate the current of each individual battery
            for battery_id in active_local_battery_states.keys():
                active_local_battery_states[battery_id].current = self.__get_battery_current(
                    active_local_battery_states, battery_id, current)
                # sort back
                local_battery_states[battery_id].current = active_local_battery_states[battery_id].current
            # Second update the battery states and check for integrity
            for battery_id in local_battery_states.keys():
                battery: LithiumIonBattery = self.__battery_dict[battery_id]
                # battery_id: int = battery.id
                battery_state: LithiumIonState = battery.get_equilibrium_state_for(time, local_battery_states[
                    battery.id].current, fixed_values=True)
                # if (abs(local_battery_states[
                #             battery_id].current - battery_state.current) > self.__ACCURACY) and battery_id in active_local_battery_states:
                if (abs(battery_state.current) < self.__CURRENT_ACCURACY) and battery_id in active_local_battery_states:
                    battery_state.current = 0
                    del active_local_battery_states[battery_id]
                    local_battery_states[battery_id] = battery_state
                    self.__log.warn('Battery ' + str(battery_id) + ' was removed from active list')
                    # if bs.current != current:
                    #     self.__log.error(
                    #         'Current of a direct parallel battery was limited by the BMS and would therefore disconnect from DC-Link. Results unusable')
                    # break
                local_battery_states[battery_id] = battery_state
                if battery_id in active_local_battery_states:
                    active_local_battery_states[battery_id] = battery_state
                    max_voltage = max(max_voltage, battery_state.voltage)
                    min_voltage = min(min_voltage, battery_state.voltage)
            # Third see which current is possible
            current = 0
            for battery_id in local_battery_states.keys():
                battery_state: LithiumIonState = local_battery_states[battery_id]
                current += battery_state.current
            # Escape while loop if no more active batteries available
            if not active_local_battery_states:
                break
            if loop_count > 10000:
                print('active_local_battery_states: ',active_local_battery_states)
                print('max_voltage: ',max_voltage, 'min_voltage: ',min_voltage)
                raise Exception('Infinite loop')
            # print(max_voltage - min_voltage)
            # print(' ')
        self.__count_dc_switch_operations(local_battery_states, active_local_battery_states)
        self.__last_active_battery_states = active_local_battery_states
        return local_battery_states

    def __filter_battery_states_by_voltage(self, active_local_battery_states: dict, current, local_battery_states: dict):
        if current > 0:
            local_battery_states = self.__sort_by_soc(local_battery_states, True)
            active_local_battery_states = self.__sort_by_soc(active_local_battery_states, True)
            ref_min_voltage: float = self.__LARGE_POSITIVE_NUMBER
            for battery_id in local_battery_states.keys():
                ref_min_voltage = min(ref_min_voltage, local_battery_states[battery_id].voltage)
            for battery_id in local_battery_states.keys():
                if abs(ref_min_voltage - active_local_battery_states[battery_id].voltage) > self.__VOLTAGE_ACCURACY:
                    local_battery_states[battery_id].current = 0
                    del active_local_battery_states[battery_id]
        else:
            local_battery_states = self.__sort_by_soc(local_battery_states, False)
            active_local_battery_states = self.__sort_by_soc(active_local_battery_states, False)
            ref_max_voltage: float = 0.0
            for battery_id in local_battery_states.keys():
                ref_max_voltage = max(ref_max_voltage, local_battery_states[battery_id].voltage)
            for battery_id in local_battery_states.keys():
                if abs(ref_max_voltage - active_local_battery_states[battery_id].voltage) > self.__VOLTAGE_ACCURACY:
                    local_battery_states[battery_id].current = 0
                    del active_local_battery_states[battery_id]
        return active_local_battery_states, local_battery_states

    def __count_dc_switch_operations(self, local_battery_states: dict, active_local_battery_states: dict) -> None:
        last_active_battery_ids = self.__last_active_battery_states.keys()
        current_active_battery_ids = active_local_battery_states.keys()
        for battery_id in local_battery_states.keys():
            if (battery_id in current_active_battery_ids and battery_id not in last_active_battery_ids) or \
                    (battery_id not in current_active_battery_ids and battery_id in last_active_battery_ids):
                self.__dc_switch_operations[battery_id] += 1

    def distribute_and_run(self, time: float, current: float, voltage: float) -> None:
        local_battery_states: dict = self.__balance_batteries(time, current)
        for battery in self.__batteries:
            # TODO improve current distribution to batteries, interface to OparaBatt (Philipp Jocher)
            # 1st approach: equal distribution of current
            # lithium_ion.set(time, current / size)
            # 2nd approach: current distribution is voltage dependent (OCV + RINT)
            battery_state: LithiumIonState = local_battery_states[battery.id]
            battery.distribute_and_run(time, battery_state.current, battery_state.voltage)
            # 3rd approach: Jaehyung Lee, A novel li-ion lithium_ion pack modeling considerging single cell
            #     information and capacity variation, https://doi.org/10.1109/ECCE.2017.8096880

    def wait(self) -> None:
        pass

    def get_auxiliaries(self) -> [Auxiliary]:
        auxiliaries: [Auxiliary] = list()
        for battery in self.__batteries:
            auxiliaries.extend(battery.get_auxiliaries())
        return auxiliaries

    @property
    def volume(self) -> float:
        volume = 0
        for battery in self.__batteries:
            volume += battery.volume
        return volume

    @property
    def mass(self) -> float:
        mass = 0
        for battery in self.__batteries:
            mass += battery.mass
        return mass

    @property
    def surface_area(self) -> float:
        surface_area = 0
        for battery in self.__batteries:
            surface_area += battery.surface_area
        return surface_area

    @property
    def specific_heat(self) -> float:
        specific_heat = 0
        for battery in self.__batteries:
            specific_heat += battery.specific_heat
        return specific_heat / len(self.__batteries)

    @property
    def convection_coefficient(self) -> float:
        convection_coefficient = 0
        for battery in self.__batteries:
            convection_coefficient += battery.convection_coefficient
        return convection_coefficient / len(self.__batteries)

    def get_system_parameters(self) -> dict:
        parameters: dict = dict()
        batteries: list = list()
        for battery in self.__batteries:
            batteries.append(battery.get_system_parameters())
        parameters[SystemParameters.BATTERIES] = batteries
        return parameters

    @property
    def state(self) -> LithiumIonState:
        battery_states: [LithiumIonState] = list()
        for battery in self.__batteries:
            battery_states.append(battery.state)
        return LithiumIonState.sum_parallel(battery_states)

    def close(self) -> None:
        """Closing all resources in lithium_ion circuit"""
        print("DC switch operations for parallel connected batteries", self.__dc_switch_operations)
        self.__log.close()
        for battery in self.__batteries:
            battery.close()



from configparser import ConfigParser
import pytest
from simses.commons.config.simulation.system import StorageSystemConfig
from simses.system.thermal.ambient.constant_temperature import ConstantAmbientTemperature

constant_ambient_temperature_local_value = 25  # in C

# Ensure if path from simulation.defaults.ini to ambient temperature model works
ambient_temperature_model_config: ConfigParser = ConfigParser()
ambient_temperature_model_config.add_section('STORAGE_SYSTEM')
ambient_temperature_model_config.set('STORAGE_SYSTEM', 'AMBIENT_TEMPERATURE_MODEL', 'ConstantAmbientTemperature,35')
storage_system_config = StorageSystemConfig(ambient_temperature_model_config)
constant_ambient_temperature_config_value = float(storage_system_config.ambient_temperature_model[StorageSystemConfig.AMBIENT_TEMPERATURE_CONSTANT])

start_time = 0
end_time = 10
sample_time = 1
time_step_range = range(start_time, end_time, sample_time)

constant_temperature_values = [constant_ambient_temperature_local_value, constant_ambient_temperature_config_value]


@pytest.mark.parametrize("time_step", time_step_range)
def test_get_temperature(time_step):
    for value in constant_temperature_values:
        uut = ConstantAmbientTemperature(value)
        assert uut.get_temperature(time_step) == value + 273.15
        assert uut.get_initial_temperature() == value + 273.15



import os
from configparser import ConfigParser
from datetime import datetime, timezone

import pytest
import time

from simses.commons.config.simulation.general import GeneralSimulationConfig
from simses.commons.config.simulation.profile import ProfileConfig
from simses.commons.config.simulation.energy_management import EnergyManagementConfig
from simses.commons.profile.technical.soc import SocProfile
from simses.commons.profile.technical.binary import BinaryProfile
from simses.commons.state.system import SystemState
from simses.commons.utils.utilities import remove_file
from simses.logic.energy_management.strategy.basic.electric_vehicle_soc import \
    ElectricVehicleSOC
from pathlib import Path

DELIMITER: str = ','
HEADER: str = '# Unit: W'
system_state: SystemState = SystemState(0,0)
START: int = 0
END: int = 4
STEP: int = 1
FILE_NAME_SOC: str = 'mockup_power_soc_profile.csv'
FILE_NAME_BINARY: str = 'mockup_power_home_profile.csv'
PATH_NAME_BINARY = os.path.join(Path(os.path.dirname(__file__)).parent.parent, 'data\profile\\technical\\' + FILE_NAME_BINARY)

def create_soc_file():
    name = os.path.join(Path(os.path.dirname(__file__)).parent.parent, 'data\profile\\technical\\' + FILE_NAME_SOC)

    with open(name, mode='w') as file:
        file.write(HEADER + '\n')
        for counter in range(END+1):
            file.write(str(counter) + DELIMITER + str(SOC_VALUES[counter]) + '\n')

    power_load_profile = SocProfile(config=create_general_config(), profile_config=create_profile_config())
    return power_load_profile

def create_binary_file():
    name = os.path.join(Path(os.path.dirname(__file__)).parent.parent, 'data\profile\\technical\\' + FILE_NAME_BINARY)

    with open(name, mode='w') as file:
        file.write(HEADER + '\n')
        for counter in range(END + 1):
            file.write(str(counter) + DELIMITER + str(BINARY_VALUES[counter]) + '\n')

    binary_profile = BinaryProfile(config=create_general_config(), profile_config=create_profile_config())
    return binary_profile

def create_general_config():
    conf: ConfigParser = ConfigParser()
    conf.add_section('GENERAL')
    date = datetime.fromtimestamp(START, timezone.utc)
    conf.set('GENERAL', 'START', date.strftime("%Y-%m-%d %H:%M:%S"))
    date = datetime.fromtimestamp(END, timezone.utc)
    conf.set('GENERAL', 'END', date.strftime("%Y-%m-%d %H:%M:%S"))
    conf.set('GENERAL', 'TIME_STEP', str(STEP))
    return GeneralSimulationConfig(config=conf)

def create_profile_config():
    conf: ConfigParser = ConfigParser()
    conf.add_section('PROFILE')
    conf.set('PROFILE', 'SOC_PROFILE', FILE_NAME_SOC)
    conf.set('PROFILE', 'BINARY_PROFILE', FILE_NAME_BINARY)
    return ProfileConfig(config=conf)

def create_ems_config():
    conf: ConfigParser = ConfigParser()
    conf.add_section('ENERGY_MANAGEMENT')
    conf.set('ENERGY_MANAGEMENT', 'STRATEGY', 'ElectricVehicleSOC')
    conf.set('ENERGY_MANAGEMENT', 'EV_CHARGING_STRATEGY', EV_CHARGING_STRATEGY)
    conf.set('ENERGY_MANAGEMENT', 'Max_Power', str(MAX_AC_POWER))
    return EnergyManagementConfig(config=conf)

@pytest.fixture(scope='module')
def uut():
    soc_profile = create_soc_file()
    binary_profile = create_binary_file()
    ems_config = create_ems_config()
    general_config = create_general_config()
    profile_config = create_profile_config()
    uut = ElectricVehicleSOC(general_config, profile_config, ems_config)

    yield uut
    soc_profile.close()
    binary_profile.close()
    uut.close()

    # remove_file(soc_profile._SocProfile__file._FileProfile__filename)
    # remove_file(binary_profile._BinaryProfile__file._FileProfile__filename)


SOC_VALUES: list = [1, 0.4, 0.5, 0.5, 0.5]
BINARY_VALUES: list = [2, 0, 1, 1, 1]
MAX_AC_POWER: int = 17
EV_CHARGING_STRATEGY: str = 'Uncontrolled'

@pytest.mark.parametrize('time, soc_state, capacity, result',
                         [
                            (0, 0.5, 10, 'ValueError'),     # 1: Binary Profil different value than 0 or 1
                            (1, 0.5, 10, -3600),            # 2: EV on road, capacity=10, SOC_state=0.5, SOC_new=0.4 => Diff=0.1; =>-3600Ws => -3600W
                            (2, 1, 10, 0),                  # 3: EV parked but SOC=100% => 0W
                            (3, 0, 10, 17),                 # 4: EV parked and SOC<100% => charged with 17W
                         ]
                         )
def test_next(time, soc_state, capacity, result, uut):

    system_state.soc = soc_state
    system_state.capacity = capacity
    if time==0:
        with pytest.raises(ValueError):
            uut.next(time, system_state)
    else:
        assert abs(uut.next(time, system_state) - result) <= 1e-10


from simses.commons.state.technology.storage import StorageTechnologyState


class ElectrolyzerState(StorageTechnologyState):
    """
    Current physical state of the electrolyzer components with the main electrical parameters.
    """

    SYSTEM_AC_ID = 'StorageSystemAC'
    SYSTEM_DC_ID = 'StorageSystemDC'
    VOLTAGE = 'voltage of electrolyzer in V'
    CURRENT = 'current of electrolyzer in A'
    CURRENT_DENSITY = 'current density of electrolyzer in A'
    POWER = 'power in W'
    TEMPERATURE = 'temperature of electrolyzer stack in K'
    POWER_LOSS = 'Power loss in W'
    HYDROGEN_PRODUCTION = 'hydrogen production in mol/s'  # hydrogen production of electrolyzerstack
    HYDROGEN_OUTFLOW = 'hydrogen outlfow in mol/s'  # hydrogen outflow after pressure controll valve
    OXYGEN_PRODUCTION = 'oxygen production in mol/s'  # oxygen production of electrolyzer
    OXYGEN_OUTFLOW = 'oxygen outflow in mol/s'  # oxygen outflow out of the electrolyzer after pressure control
    FULFILLMENT = 'fulfillment in p.u.'  # ??
    CONVECTION_HEAT = 'convection heat in W'
    PRESSURE_ANODE = 'relative anode pressure of electrolyzer in barg'
    PRESSURE_CATHODE = 'relative cathode pressure or electrolyzer in barg'
    PART_PRESSURE_H2 = 'partial pressure H2 in bar'
    PART_PRESSURE_O2 = 'partial pressure O2 in bar'
    SAT_PRESSURE_H2O = 'saturation pressure H2O in bar'
    WATER_USE = 'wateruse in mol/s'  # water the electrolyzer splits into hydrogen and oxygen
    WATER_OUTFLOW_CATHODE = 'watersteam outflow cathode electrolyzer in mol/s'
    WATER_OUTFLOW_ANODE = 'watersteam outflow anode electrolyzer in mol/s'
    WATER_FLOW = 'waterflow electrolyzer in mol/s'  # water the is circulated through the EL stack in order to tempering
    POWER_WATER_HEATING  = 'power water heating electrolyzer in W'  # power that is transported by the water stream through the electrolyzer stack
                                                                        # if Ph20 > 0: heating  if Ph2o < 0: heat is transported out of the stack and realeased to the ambient atmosphere (no electrical power needed)
    POWER_PUMP = 'power water circulation electrolyzer in W'  # power the pump consumes for circulation of water through electrolyzer
    POWER_GAS_DRYING = 'power for drying of hydrogen in W'
    POWER_COMPRESSOR = 'power for compression of hydrogen in W'
    TOTAL_HYDROGEN_PRODUCTION = 'total amount of produced hydrogen in kg'
    RESISTANCE_INCREASE_CYCLIC = 'increase resistance cyclic in p.u.'
    RESISTANCE_INCREASE_CALENDAR = 'increase reisistance calendric in p.u.'
    RESISTANCE_INCREASE = 'increase R total in p.u.'
    REFERENCE_VOLTAGE = 'reverence voltage in V'
    EXCHANGE_CURRENT_DENS_DECREASE_CYCLIC = 'decrease j0 cyclic in p.u.'
    EXCHANGE_CURRENT_DENS_DECREASE_CALENDAR = 'decrease j0 calendric in p.u.'
    EXCHANGE_CURRENT_DENS_DECREASE = 'decrease j0 in p.u.'
    SOH = 'SOH electrolyzer in p.u.'
    MAX_CHARGE_POWER = 'Maximum charging power in W'
    MAX_DISCHARGE_POWER = 'Maximum discharging power in W'

    def __init__(self, system_id: int, storage_id: int):
        super().__init__()
        self._initialize()
        self.set(self.SYSTEM_AC_ID, system_id)
        self.set(self.SYSTEM_DC_ID, storage_id)

    @property
    def soc(self) -> float:
        pass

    @soc.setter
    def soc(self, value: float) -> None:
        pass

    @property
    def capacity(self) -> float:
        pass

    @capacity.setter
    def capacity(self, value: float):
        pass

    @property
    def voltage(self) -> float:
        return self.get(self.VOLTAGE)

    @voltage.setter
    def voltage(self, value: float) -> None:
        self.set(self.VOLTAGE, value)

    @property
    def current(self) -> float:
        return self.get(self.CURRENT)

    @current.setter
    def current(self, value: float) -> None:
        self.set(self.CURRENT, value)

    @property
    def current_density(self) -> float:
        return self.get(self.CURRENT_DENSITY)

    @current_density.setter
    def current_density(self, value: float) -> None:
        self.set(self.CURRENT_DENSITY, value)

    @property
    def temperature(self) -> float:
        return self.get(self.TEMPERATURE)

    @temperature.setter
    def temperature(self, value: float) -> None:
        self.set(self.TEMPERATURE, value)

    @property
    def power_loss(self) -> float:
        return self.get(self.POWER_LOSS)

    @power_loss.setter
    def power_loss(self, value: float) -> None:
        self.set(self.POWER_LOSS, value)

    @property
    def power(self) -> float:
        return self.get(self.POWER)

    @power.setter
    def power(self, value: float) -> None:
        self.set(self.POWER, value)

    @property
    def hydrogen_production(self) -> float:
        return self.get(self.HYDROGEN_PRODUCTION)

    @hydrogen_production.setter
    def hydrogen_production(self, value: float) -> None:
        self.set(self.HYDROGEN_PRODUCTION, value)

    @property
    def hydrogen_outflow(self) -> float:
        return self.get(self.HYDROGEN_OUTFLOW)

    @hydrogen_outflow.setter
    def hydrogen_outflow(self, value: float) -> None:
        self.set(self.HYDROGEN_OUTFLOW, value)

    @property
    def oxygen_production(self) -> float:
        return self.get(self.OXYGEN_PRODUCTION)

    @oxygen_production.setter
    def oxygen_production(self, value: float) -> None:
        self.set(self.OXYGEN_PRODUCTION, value)

    @property
    def oxygen_outflow(self) -> float:
        return self.get(self.OXYGEN_OUTFLOW)

    @oxygen_outflow.setter
    def oxygen_outflow(self, value: float) -> None:
        self.set(self.OXYGEN_OUTFLOW, value)

    @property
    def convection_heat(self) -> float:
        return self.get(self.CONVECTION_HEAT)

    @convection_heat.setter
    def convection_heat(self, value: float) -> None:
        self.set(self.CONVECTION_HEAT, value)

    @property
    def fulfillment(self) -> float:
        return self.get(self.FULFILLMENT)

    @fulfillment.setter
    def fulfillment(self, value: float):
        self.set(self.FULFILLMENT, value)

    @property
    def pressure_anode(self) -> float:
        return self.get(self.PRESSURE_ANODE)

    @pressure_anode.setter
    def pressure_anode(self, value: float):
        self.set(self.PRESSURE_ANODE, value)

    @property
    def pressure_cathode(self) -> float:
        return self.get(self.PRESSURE_CATHODE)

    @pressure_cathode.setter
    def pressure_cathode(self, value: float):
        self.set(self.PRESSURE_CATHODE, value)

    @property
    def part_pressure_h2(self) -> float:
        return self.get(self.PART_PRESSURE_H2)

    @part_pressure_h2.setter
    def part_pressure_h2(self, value: float):
        self.set(self.PART_PRESSURE_H2, value)

    @property
    def part_pressure_o2(self) -> float:
        return self.get(self.PART_PRESSURE_O2)

    @part_pressure_o2.setter
    def part_pressure_o2(self, value: float):
        self.set(self.PART_PRESSURE_O2, value)

    @property
    def sat_pressure_h2o(self) -> float:
        return self.get(self.SAT_PRESSURE_H2O)

    @sat_pressure_h2o.setter
    def sat_pressure_h2o(self, value: float):
        self.set(self.SAT_PRESSURE_H2O, value)

    @property
    def water_use(self) -> float:
        return self.get(self.WATER_USE)

    @water_use.setter
    def water_use(self, value: float):
        self.set(self.WATER_USE, value)

    @property
    def water_outflow_cathode(self) -> float:
        return self.get(self.WATER_OUTFLOW_CATHODE)

    @water_outflow_cathode.setter
    def water_outflow_cathode(self, value: float):
        self.set(self.WATER_OUTFLOW_CATHODE, value)

    @property
    def water_outflow_anode(self) -> float:
        return self.get(self.WATER_OUTFLOW_ANODE)

    @water_outflow_anode.setter
    def water_outflow_anode(self, value: float):
        self.set(self.WATER_OUTFLOW_ANODE, value)

    @property
    def water_flow(self) -> float:
        return self.get(self.WATER_FLOW)

    @water_flow.setter
    def water_flow(self, value: float):
        self.set(self.WATER_FLOW, value)

    @property
    def power_water_heating(self) -> float:
        return self.get(self.POWER_WATER_HEATING)

    @power_water_heating.setter
    def power_water_heating(self, value: float):
        self.set(self.POWER_WATER_HEATING, value)

    @property
    def power_pump(self) -> float:
        return self.get(self.POWER_PUMP)

    @power_pump.setter
    def power_pump(self, value: float):
        self.set(self.POWER_PUMP, value)

    @property
    def power_gas_drying(self) -> float:
        return self.get(self.POWER_GAS_DRYING)

    @power_gas_drying.setter
    def power_gas_drying(self, value: float):
        self.set(self.POWER_GAS_DRYING, value)

    @property
    def power_compressor(self) -> float:
        return self.get(self.POWER_COMPRESSOR)

    @power_compressor.setter
    def power_compressor(self, value: float):
        self.set(self.POWER_COMPRESSOR, value)

    @property
    def total_hydrogen_production(self) -> float:
        return self.get(self.TOTAL_HYDROGEN_PRODUCTION)

    @total_hydrogen_production.setter
    def total_hydrogen_production(self, value: float):
        self.set(self.TOTAL_HYDROGEN_PRODUCTION, value)

    @property
    def resistance_increase_cyclic(self) -> float:
        return self.get(self.RESISTANCE_INCREASE_CYCLIC)

    @resistance_increase_cyclic.setter
    def resistance_increase_cyclic(self, value: float):
        self.set(self.RESISTANCE_INCREASE_CYCLIC, value)

    @property
    def resistance_increase_calendar(self) -> float:
        return self.get(self.RESISTANCE_INCREASE_CALENDAR)

    @resistance_increase_calendar.setter
    def resistance_increase_calendar(self, value: float):
        self.set(self.RESISTANCE_INCREASE_CALENDAR, value)

    @property
    def resistance_increase(self) -> float:
        return self.get(self.RESISTANCE_INCREASE)

    @resistance_increase.setter
    def resistance_increase(self, value: float):
        self.set(self.RESISTANCE_INCREASE, value)

    @property
    def reference_voltage(self) -> float:
        return self.get(self.REFERENCE_VOLTAGE)

    @reference_voltage.setter
    def reference_voltage(self, value: float):
        self.set(self.REFERENCE_VOLTAGE, value)

    @property
    def exchange_current_decrease_cyclic(self) -> float:
        return self.get(self.EXCHANGE_CURRENT_DENS_DECREASE_CYCLIC)

    @exchange_current_decrease_cyclic.setter
    def exchange_current_decrease_cyclic(self, value: float):
        self.set(self.EXCHANGE_CURRENT_DENS_DECREASE_CYCLIC, value)

    @property
    def exchange_current_decrease_calendar(self) -> float:
        return self.get(self.EXCHANGE_CURRENT_DENS_DECREASE_CALENDAR)

    @exchange_current_decrease_calendar.setter
    def exchange_current_decrease_calendar(self, value: float):
        self.set(self.EXCHANGE_CURRENT_DENS_DECREASE_CALENDAR, value)

    @property
    def exchange_current_decrease(self) -> float:
        return self.get(self.EXCHANGE_CURRENT_DENS_DECREASE)

    @exchange_current_decrease.setter
    def exchange_current_decrease(self, value: float):
        self.set(self.EXCHANGE_CURRENT_DENS_DECREASE, value)

    @property
    def max_charge_power(self) -> float:
        return self.get(self.MAX_CHARGE_POWER)

    @max_charge_power.setter
    def max_charge_power(self, value: float) -> None:
        self.set(self.MAX_CHARGE_POWER, value)

    @property
    def max_discharge_power(self) -> float:
        return self.get(self.MAX_DISCHARGE_POWER)

    @max_discharge_power.setter
    def max_discharge_power(self, value: float) -> None:
        self.set(self.MAX_DISCHARGE_POWER, value)

    @property
    def soh(self) -> float:
        return self.get(self.SOH)

    @soh.setter
    def soh(self, value: float):
        self.set(self.SOH, value)

    @property
    def id(self) -> str:
        return 'ELECTROLYZER' + str(self.get(self.SYSTEM_AC_ID)) + str(self.get(self.SYSTEM_DC_ID))

    @property
    def is_charge(self) -> bool:
        return self.power > 0

    @classmethod
    def sum_parallel(cls, hydrogen_states: []):
        state = ElectrolyzerState(0, 0)
        return state

    @classmethod
    def sum_serial(cls, states: []):
        pass
